from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from tests.integration.io.conftest import minio_create_bucket

TABLE_NAME = "my_table"
data1 = {
    "vector": [[1.1, 1.2], [0.2, 1.8]],
    "lat": [45.5, 40.1],
    "long": [-122.7, -74.1],
}

data2 = {
    "vector": [[2.1, 2.2], [2.2, 2.8]],
    "lat": [46.5, 41.1],
    "long": [-123.7, -75.1],
}

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(PYARROW_LOWER_BOUND_SKIP, reason="lance not supported on old versions of pyarrow")


@pytest.fixture(scope="function")
def lance_dataset_path(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("lance")
    yield str(tmp_dir)


def test_lancedb_roundtrip(lance_dataset_path):
    df1 = daft.from_pydict(data1)
    df2 = daft.from_pydict(data2)
    df1.write_lance(lance_dataset_path, mode="create")
    df2.write_lance(lance_dataset_path, mode="append")
    df_loaded = daft.read_lance(lance_dataset_path)

    assert df_loaded.to_pydict() == df1.concat(df2).to_pydict()


@pytest.mark.integration()
def test_lancedb_minio(minio_io_config):
    df1 = daft.from_pydict(data1)
    df2 = daft.from_pydict(data2)
    bucket_name = "lance"
    s3_path = f"s3://{bucket_name}/data"
    with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        df1.write_lance(s3_path, mode="create", io_config=minio_io_config)
        df2.write_lance(s3_path, mode="append", io_config=minio_io_config)
        df_loaded = daft.read_lance(s3_path, io_config=minio_io_config)
        assert df_loaded.to_pydict() == df1.concat(df2).to_pydict()


def test_lancedb_write_with_schema(lance_dataset_path):
    """Writing a dataframe to lance with a user-provided schema with lance encodings."""
    data = {
        "vector": [1.1, 1.2],
        "compressible_strings": ["a" * 100, "b" * 100],
    }
    df = daft.from_pydict(data)

    pa_schema = pa.schema(
        [
            pa.field("vector", pa.float64()),
            pa.field("compressible_strings", pa.string(), metadata={"lance-encoding:compression": "zstd"}),
        ]
    )
    daft_schema = daft.schema.Schema.from_pyarrow_schema(pa_schema)
    print(f"\n daft schema:\n {daft_schema}; \n with metadata: \n{daft_schema.display_with_metadata(True)}")
    assert daft_schema.__repr__() == daft_schema.display_with_metadata(False) + "\n"
    df.write_lance(lance_dataset_path, schema=daft_schema, mode="create")

    df_loaded = daft.read_lance(lance_dataset_path)
    assert df_loaded.to_pydict() == df.to_pydict()

    import lance

    ds = lance.dataset(lance_dataset_path)
    written_pa_schema = ds.schema

    compress_field = written_pa_schema.field("compressible_strings")
    assert compress_field.type == pa.large_string()

    compress_field_metadata = compress_field.metadata
    assert compress_field_metadata is not None
    assert compress_field_metadata[b"lance-encoding:compression"] == b"zstd"



"""This module tests examples from https://pandas.pydata.org/docs/user_guide/cookbook.html."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import daft
from daft.datatype import DataType
from daft.expressions import col, lit
from tests.conftest import assert_df_equals

###
# Idioms: if-then
###

IF_THEN_DATA = {"AAA": [4, 5, 6, 7], "BBB": [10, 20, 30, 40], "CCC": [100, 50, -30, -50]}


def test_if_then(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(IF_THEN_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(IF_THEN_DATA)
    daft_df = daft_df.with_column("BBB", (col("AAA") >= 5).if_else(-1, col("BBB")))
    pd_df.loc[pd_df.AAA >= 5, "BBB"] = -1
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


def test_if_then_2_cols(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(IF_THEN_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(IF_THEN_DATA)
    daft_df = daft_df.with_column("BBB", (col("AAA") >= 5).if_else(2000, col("BBB"))).with_column(
        "CCC",
        (col("AAA") >= 5).if_else(2000, col("CCC")),
    )
    pd_df.loc[pd_df.AAA >= 5, ["BBB", "CCC"]] = 2000
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


def test_if_then_numpy_where(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(IF_THEN_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(IF_THEN_DATA)
    daft_df = daft_df.with_column("logic", (col("AAA") > 5).if_else("high", lit("low")))
    pd_df["logic"] = np.where(pd_df["AAA"] > 5, "high", "low")
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


###
# Idioms: splitting
###

SPLITTING_DATA = {"AAA": [4, 5, 6, 7], "BBB": [10, 20, 30, 40], "CCC": [100, 50, -30, -50]}


def test_split_frame_boolean_criterion(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(SPLITTING_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(SPLITTING_DATA)
    daft_df = daft_df.where(col("AAA") <= 5)
    pd_df = pd_df[pd_df.AAA <= 5]
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


###
# Idioms: Building criteria
###

BUILDING_DATA = {"AAA": [4, 5, 6, 7], "BBB": [10, 20, 30, 40], "CCC": [100, 50, -30, -50]}


def test_multi_criteria_and(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(BUILDING_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(BUILDING_DATA)
    daft_df = daft_df.where((col("BBB") < 25) & (col("CCC") >= -40)).select(col("AAA"))
    pd_df = pd.DataFrame({"AAA": pd_df.loc[(pd_df["BBB"] < 25) & (pd_df["CCC"] >= -40), "AAA"]})
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


def test_multi_criteria_or(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(BUILDING_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(BUILDING_DATA)
    daft_df = daft_df.where((col("BBB") > 25) | (col("CCC") >= -40)).select(col("AAA"))
    pd_df = pd.DataFrame({"AAA": pd_df.loc[(pd_df["BBB"] > 25) | (pd_df["CCC"] >= -40), "AAA"]})
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


def test_multi_criteria_or_assignment(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(BUILDING_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(BUILDING_DATA)
    daft_df = daft_df.with_column(
        "AAA", ((col("BBB") > 25) | (col("CCC") >= 75)).if_else(0.1, col("AAA").cast(DataType.float32()))
    )
    pd_df.loc[(pd_df["BBB"] > 25) | (pd_df["CCC"] >= 75), "AAA"] = 0.1
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="BBB")


def test_select_rows_closest_to_certain_value_using_argsort(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(BUILDING_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(BUILDING_DATA)
    aValue = 43.0
    daft_df = daft_df.sort(abs(col("CCC") - aValue))
    pd_df = pd_df.loc[(pd_df.CCC - aValue).abs().argsort()]
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


###
# Selection: Dataframes
###

SELECTION_DATA = {"AAA": [4, 5, 6, 7], "BBB": [10, 20, 30, 40], "CCC": [100, 50, -30, -50]}


@pytest.mark.skip(reason="Requires F.row_number() and Expression.is_in(...)")
def test_splitting_by_row_index(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(SELECTION_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(SELECTION_DATA)
    daft_df = daft_df.where((col("AAA") <= 6) & F.row_number().is_in([0, 2, 4]))  # noqa: F821
    pd_df = pd_df[(pd_df.AAA <= 6) & (pd_df.index.isin([0, 2, 4]))]
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


@pytest.mark.skip(reason="Requires F.row_number()")
def test_splitting_by_row_range(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(SELECTION_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(SELECTION_DATA)
    daft_df = daft_df.where((F.row_number() >= 0) & (F.row_number() < 3))  # noqa: F821
    pd_df = pd_df[0:3]
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


###
# Selection: New Columns
###

APPLYMAP_DATA = {"AAA": [1, 2, 1, 3], "BBB": [1, 1, 2, 2], "CCC": [2, 1, 3, 1]}


@pytest.mark.skip(reason="Requires Expression.applymap((val) => result)")
def test_efficiently_and_dynamically_creating_new_columns_using_applymap(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(APPLYMAP_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(APPLYMAP_DATA)
    source_cols = pd_df.columns
    categories = {1: "Alpha", 2: "Beta", 3: "Charlie"}
    new_cols = [str(x) + "_cat" for x in source_cols]

    for source_col, new_col in zip(source_cols, new_cols):
        daft_df = daft_df.with_column(new_col, col(source_col).applymap(lambda source_val: categories.get(source_val)))

    pd_df[new_cols] = pd_df[source_cols].applymap(categories.get)

    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


MIN_WITH_GROUPBY_DATA = {"AAA": [1, 1, 1, 2, 2, 2, 3, 3], "BBB": [2, 1, 3, 4, 5, 1, 2, 3]}


@pytest.mark.skip(reason="Requires .first() aggregations")
def test_keep_other_columns_when_using_min_with_groupby(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(APPLYMAP_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(APPLYMAP_DATA)
    daft_df = daft_df.groupby(col("AAA")).min(col("BBB"))
    pd_df = pd_df.sort_values(by="BBB").groupby("AAA", as_index=False).first()
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="AAA")


###
# Missing Data: We do not support custom utilities for missing data at this time.
###

###
# Grouping: We do not support
#   1. .apply
#   2. .get_group
#   3. .transform
# Many of the tests in the Pandas cookbook are skipped here
###

GROUPBY_DATA = {
    "animal": "cat dog cat fish dog cat cat".split(),
    "size": list("SSMMMLL"),
    "weight": [8, 10, 11, 1, 20, 12, 12],
    "adult": [False] * 5 + [True] * 2,
}

# TODO: We can't do this yet, since it requires getting the first row for every Group, ordered by some column
# @pytest.mark.tdd
# @parametrize_partitioned_daft_df(source=GROUPBY_DATA, partitioning=GROUPBY_DATA_PARTITIONING)
# def test_basic_grouping_with_apply(repartition_nparts):
#     """Gets the sizes of the animals with the highest weight"""
#     pd_df = pd.DataFrame(pd_df.groupby("animal").apply(lambda subf: subf["size"][subf["weight"].idxmax()]), columns="size").reset_index()
#     assert_df_equals(daft_df, pd_df, sort_key="animal")


def test_applying_to_different_items_in_group(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(GROUPBY_DATA).repartition(repartition_nparts)
    pd_df = pd.DataFrame.from_dict(GROUPBY_DATA)
    daft_df = daft_df.with_column(
        "weight", (col("size") == "S").if_else(col("weight") * 1.5, col("weight").cast(DataType.float32()))
    )
    daft_df = daft_df.with_column("weight", (col("size") == "M").if_else(col("weight") * 1.25, col("weight")))
    daft_df = daft_df.with_column("weight", (col("size") == "L").if_else(col("weight"), col("weight")))
    daft_df = daft_df.groupby(col("animal")).agg(col("weight").mean())
    daft_df = daft_df.with_column("size", lit("L"))
    daft_df = daft_df.with_column("adult", lit(True))

    def GrowUp(x):
        avg_weight = sum(x[x["size"] == "S"].weight * 1.5)
        avg_weight += sum(x[x["size"] == "M"].weight * 1.25)
        avg_weight += sum(x[x["size"] == "L"].weight)
        avg_weight /= len(x)
        return pd.Series(["L", avg_weight, True], index=["size", "weight", "adult"])

    pd_df = pd_df.groupby("animal").apply(GrowUp).reset_index()
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="animal")


###
# Merge
###

JOIN_DATA = {
    "Area": ["A"] * 5 + ["C"] * 2,
    "Bins": [110] * 2 + [160] * 3 + [40] * 2,
    "Test_0": [0, 1, 0, 1, 2, 0, 1],
    "Data": np.random.randn(7),
}


def test_self_join(repartition_nparts, with_morsel_size):
    daft_df = daft.from_pydict(JOIN_DATA).repartition(repartition_nparts)
    daft_df = daft_df.with_column("Test_1", col("Test_0") - 1)
    daft_df = daft_df.join(
        daft_df, left_on=[col("Bins"), col("Area"), col("Test_0")], right_on=[col("Bins"), col("Area"), col("Test_1")]
    )
    daft_pd_df = daft_df.to_pandas()

    pd_df = pd.DataFrame.from_dict(JOIN_DATA)
    pd_df["Test_1"] = pd_df["Test_0"] - 1
    pd_df = pd.merge(
        pd_df,
        pd_df,
        left_on=["Bins", "Area", "Test_0"],
        right_on=["Bins", "Area", "Test_1"],
        suffixes=("", "_R"),
    )
    pd_df = pd_df.rename({"Test_0_R": "right.Test_0", "Data_R": "right.Data", "Test_1_R": "right.Test_1"}, axis=1)
    #   Area  Bins  Test_0_L    Data_L  Test_1_L  Test_0_R    Data_R  Test_1_R
    # 0    A   110         0 -0.433937        -1         1 -0.160552         0
    # 1    A   160         0  0.744434        -1         1  1.754213         0
    # 2    A   160         1  1.754213         0         2  0.000850         1
    # 3    C    40         0  0.342243        -1         1  1.070599         0
    assert_df_equals(daft_pd_df, pd_df, sort_key="Data")



from __future__ import annotations

from typing import Any

import pytest

import daft
from daft import col
from daft.datatype import DataTypeLike


@pytest.fixture(scope="function")
def deserialize():
    def _deserialize(items: list[str], format: str, dtype: DataTypeLike) -> list[Any]:
        field_name = "field_to_deserialize"
        df = daft.from_pydict({field_name: items})
        df = df.select(col(field_name).deserialize(format, dtype))  # type: ignore
        return df.to_pydict()[field_name]

    yield _deserialize


@pytest.fixture(scope="function")
def try_deserialize():
    def _try_deserialize(items: list[str], format: str, dtype: DataTypeLike) -> list[Any]:
        field_name = "field_to_deserialize"
        df = daft.from_pydict({field_name: items})
        df = df.select(col(field_name).deserialize(format, dtype))  # type: ignore
        return df.to_pydict()[field_name]

    yield _try_deserialize



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pac
import pytest
import ray

import daft
from daft import DataType
from tests.conftest import get_tests_daft_runner_name

RAY_VERSION = tuple(int(s) for s in ray.__version__.split(".")[0:3])


class MyObj:
    def __init__(self, x: int):
        self._x = x

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MyObj) and self._x == other._x


DATA = {
    "intcol": [1, 2, 3],
    "strcol": ["a", "b", "c"],
}


def _row_to_pydict(row: ray.data.row.TableRow | dict) -> dict:
    if isinstance(row, dict):
        return row
    return row.as_pydict()


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_ray_dataset_all_arrow(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("floatcol", df["intcol"].cast(DataType.float64()))
    ds = df.to_ray_dataset()

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert ds.dataset_format() == "arrow", "Ray Dataset format should be arrow"
        elif RAY_VERSION >= (2, 0, 0):
            assert ds._dataset_format() == "arrow", "Ray Dataset format should be arrow"

    rows = sorted([_row_to_pydict(row) for row in ds.iter_rows()], key=lambda r: r["intcol"])
    assert rows == sorted(
        [
            {"intcol": 1, "strcol": "a", "floatcol": 1.0},
            {"intcol": 2, "strcol": "b", "floatcol": 2.0},
            {"intcol": 3, "strcol": "c", "floatcol": 3.0},
        ],
        key=lambda r: r["intcol"],
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.skipif(
    RAY_VERSION >= (2, 5, 0), reason="Ray Datasets versions >= 2.5.0 no longer support Python objects as rows"
)
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_ray_dataset_with_py(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("pycol", df["intcol"].apply(lambda x: MyObj(x), DataType.python()))
    ds = df.to_ray_dataset()

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert ds.dataset_format() == "simple", "Ray Dataset format should be simple because it has Python objects"
        elif RAY_VERSION >= (2, 0, 0):
            assert ds._dataset_format() == "simple", "Ray Dataset format should be simple because it has Python objects"

    rows = sorted([row for row in ds.iter_rows()], key=lambda r: r["intcol"])
    assert rows == sorted(
        [
            {"intcol": 1, "strcol": "a", "pycol": MyObj(1)},
            {"intcol": 2, "strcol": "b", "pycol": MyObj(2)},
            {"intcol": 3, "strcol": "c", "pycol": MyObj(3)},
        ],
        key=lambda r: r["intcol"],
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_ray_dataset_with_numpy(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    shape = (3, 3)
    df = df.with_column("npcol", df["intcol"].apply(lambda _: np.ones(shape), DataType.tensor(DataType.int64(), shape)))
    ds = df.to_ray_dataset()

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert (
                ds.dataset_format() == "arrow"
            ), "Ray Dataset format should be arrow because it uses a Tensor extension type"
        elif RAY_VERSION >= (2, 0, 0):
            assert (
                ds._dataset_format() == "arrow"
            ), "Ray Dataset format should be arrow because it uses a Tensor extension type"

    rows = sorted([_row_to_pydict(row) for row in ds.iter_rows()], key=lambda r: r["intcol"])
    np.testing.assert_equal(
        rows,
        sorted(
            [
                {"intcol": 1, "strcol": "a", "npcol": np.ones((3, 3))},
                {"intcol": 2, "strcol": "b", "npcol": np.ones((3, 3))},
                {"intcol": 3, "strcol": "c", "npcol": np.ones((3, 3))},
            ],
            key=lambda r: r["intcol"],
        ),
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.skipif(RAY_VERSION < (2, 2, 0), reason="Variable-shaped tensor columns not supported in Ray < 2.1.0")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_ray_dataset_with_numpy_variable_shaped(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column(
        "npcol", df["intcol"].apply(lambda x: np.ones((x, 3), dtype=np.int64), DataType.tensor(DataType.int64()))
    )
    ds = df.to_ray_dataset()

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert (
                ds.dataset_format() == "arrow"
            ), "Ray Dataset format should be arrow because it uses a Tensor extension type"
        elif RAY_VERSION >= (2, 0, 0):
            assert (
                ds._dataset_format() == "simple"
            ), "In old versions of Ray, we drop down to `simple` format because ArrowTensorType is not compatible with ragged tensors"

    rows = sorted([_row_to_pydict(row) for row in ds.iter_rows()], key=lambda r: r["intcol"])
    np.testing.assert_equal(
        rows,
        sorted(
            [
                {"intcol": 1, "strcol": "a", "npcol": np.ones((1, 3))},
                {"intcol": 2, "strcol": "b", "npcol": np.ones((2, 3))},
                {"intcol": 3, "strcol": "c", "npcol": np.ones((3, 3))},
            ],
            key=lambda r: r["intcol"],
        ),
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_all_arrow(n_partitions: int):
    def add_float(table: pa.Table) -> pa.Table:
        return table.append_column("floatcol", table["intcol"].cast(pa.float64()))

    table = pa.table(DATA)
    ds = ray.data.from_arrow(table).map_batches(add_float, batch_format="pyarrow").repartition(n_partitions)

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert ds.dataset_format() == "arrow", "Ray Dataset format should be arrow"
        elif RAY_VERSION >= (2, 0, 0):
            assert ds._dataset_format() == "arrow", "Ray Dataset format should be arrow"

    df = daft.from_ray_dataset(ds)
    # Sort data since partition ordering in Datasets is not deterministic.
    out_table = df.to_arrow()
    out_table = out_table.take(pac.sort_indices(out_table, sort_keys=[("intcol", "ascending")]))
    expected_table = add_float(table).cast(
        pa.schema(
            [
                ("intcol", pa.int64()),
                ("strcol", pa.large_string()),
                ("floatcol", pa.float64()),
            ]
        )
    )
    assert out_table.equals(expected_table), (out_table, expected_table)


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_simple(n_partitions: int):
    ds = ray.data.range(8, parallelism=n_partitions)

    df = daft.from_ray_dataset(ds)
    # Sort data since partition ordering in Datasets is not deterministic.
    out = df.to_pydict()
    key = "id" if RAY_VERSION >= (2, 5, 0) else "value"
    assert list(out.keys()) == [key]
    assert sorted(out[key]) == list(range(8))


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_tensor(n_partitions: int):
    ds = ray.data.range(8)
    ds = (
        ds.map(lambda d: {"int": d["id"], "np": np.ones((3, 3))})
        if RAY_VERSION >= (2, 5, 0)
        else ds.map(lambda i: {"int": i, "np": np.ones((3, 3))})
    )
    ds = ds.repartition(n_partitions)

    df = daft.from_ray_dataset(ds)
    out = df.to_pydict()
    assert out.keys() == {"int", "np"}
    # Sort data since partition ordering in Datasets is not deterministic.
    out_sorted_rows = sorted(list(zip(out["int"], out["np"])), key=lambda row: row[0])
    int_col, np_col = zip(*out_sorted_rows)
    out_sorted = {"int": int_col, "np": np_col}
    expected = {
        "int": list(range(8)),
        "np": [np.ones((3, 3)) for i in range(8)],
    }
    np.testing.assert_equal(out_sorted, expected)


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_pandas(n_partitions: int):
    def add_float(df: pd.DataFrame) -> pd.DataFrame:
        df["floatcol"] = df["intcol"].astype(float)
        return df

    pd_df = pd.DataFrame(DATA)

    ds = ray.data.from_pandas(pd_df).map_batches(add_float).repartition(n_partitions)

    if RAY_VERSION < (2, 4, 0):
        if RAY_VERSION >= (2, 2, 0):
            assert ds.dataset_format() == "pandas", "Ray Dataset format should be pandas"
        elif RAY_VERSION >= (2, 0, 0):
            assert ds._dataset_format() == "pandas", "Ray Dataset format should be pandas"

    df = daft.from_ray_dataset(ds)
    expected_df = add_float(pd_df)
    pd.testing.assert_frame_equal(df.to_pandas(), expected_df)


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_preview(n_partitions: int):
    ds = ray.data.range(3, parallelism=n_partitions)

    df = daft.from_ray_dataset(ds)
    assert len(df) == 3
    assert len(df._preview.partition) == 3


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_ray_dataset_data_longer_than_preview(n_partitions: int):
    ds = ray.data.range(10, parallelism=n_partitions)

    df = daft.from_ray_dataset(ds)
    assert len(df) == 10
    assert len(df._preview.partition) == 8



"""This file defines a FastAPI server that emulates an S3 implementation that serves Parquet files.

This S3 implementation serves a small Parquet file at the location: `s3://{bucket-name}/{status_code}/{num_errors}/{item_id}`

1. `status_code` is the HTTP status code it returns when S3 clients hit that endpoint
2. `num_errors` is the number of times it throws that error before successfully returning data for a given requested byte range
3. `item_id` is an ID generated by the clients to uniquely identify one attempt at retrieving this Parquet file

We provide two different buckets, with slightly different behavior:

1. "head-retries-parquet-bucket": this bucket throws errors during HEAD operations
2. "get-retries-parquet-bucket": this bucket throws errors during the ranged GET operations
"""

from __future__ import annotations

from fastapi import FastAPI

from .routers import (
    rate_limited_echo_gets_bucket,
    retries_parquet_bucket,
)

app = FastAPI()
app.mount(retries_parquet_bucket.route, retries_parquet_bucket.app)
app.mount(rate_limited_echo_gets_bucket.route, rate_limited_echo_gets_bucket.app)



from __future__ import annotations

from daft.datatype import DataType
from daft.expressions import col
from tests.expressions.typing.conftest import assert_typing_resolve_vs_runtime_behavior


def test_float_is_nan(unary_data_fixture):
    assert_typing_resolve_vs_runtime_behavior(
        data=[unary_data_fixture],
        expr=col(unary_data_fixture.name()).float.is_nan(),
        run_kernel=unary_data_fixture.float.is_nan,
        resolvable=unary_data_fixture.datatype() in (DataType.float32(), DataType.float64(), DataType.null()),
    )


def test_float_is_inf(unary_data_fixture):
    assert_typing_resolve_vs_runtime_behavior(
        data=[unary_data_fixture],
        expr=col(unary_data_fixture.name()).float.is_inf(),
        run_kernel=unary_data_fixture.float.is_inf,
        resolvable=unary_data_fixture.datatype() in (DataType.float32(), DataType.float64(), DataType.null()),
    )


def test_float_not_nan(unary_data_fixture):
    assert_typing_resolve_vs_runtime_behavior(
        data=[unary_data_fixture],
        expr=col(unary_data_fixture.name()).float.not_nan(),
        run_kernel=unary_data_fixture.float.not_nan,
        resolvable=unary_data_fixture.datatype() in (DataType.float32(), DataType.float64(), DataType.null()),
    )


def test_fill_nan(binary_data_fixture):
    lhs, rhs = binary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=binary_data_fixture,
        expr=col(lhs.name()).float.fill_nan(col(rhs.name())),
        run_kernel=lambda: lhs.float.fill_nan(rhs),
        resolvable=(
            lhs.datatype() in (DataType.float32(), DataType.float64())
            and rhs.datatype() in (DataType.float32(), DataType.float64())
        )
        or lhs.datatype() == DataType.null(),
    )



from __future__ import annotations

import contextlib
import os
import subprocess
import sys

import pytest


@contextlib.contextmanager
def with_otel_endpoint():
    old_otel_endpoint = os.getenv("DAFT_DEV_OTEL_EXPORTER_OTLP_ENDPOINT")
    os.environ["DAFT_DEV_OTEL_EXPORTER_OTLP_ENDPOINT"] = "grpc://localhost:4317"

    try:
        yield
    finally:
        if old_otel_endpoint is not None:
            os.environ["DAFT_DEV_OTEL_EXPORTER_OTLP_ENDPOINT"] = old_otel_endpoint
        else:
            del os.environ["DAFT_DEV_OTEL_EXPORTER_OTLP_ENDPOINT"]


def test_basic_otel_usage() -> None:
    pytest.skip("otel is not fully implemented")
    daft_script = """
import daft

df = daft.from_pydict({"a": [1, 2, 3, 4], "b": [1, None, 3, None]})
res = df.count("*").to_pydict()
assert res == {"count": [4]}
    """
    with with_otel_endpoint():
        result = subprocess.run([sys.executable, "-c", daft_script], capture_output=True, env=os.environ)
        assert result.stdout.decode().strip() == ""



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import Optional, Union

from daft.api_annotations import PublicAPI
from daft.context import get_context
from daft.daft import IOConfig
from daft.daft import PyRecordBatch as _PyRecordBatch
from daft.dataframe import DataFrame
from daft.logical.builder import LogicalPlanBuilder
from daft.recordbatch import MicroPartition
from daft.runners.partitioning import LocalPartitionSet


@PublicAPI
def from_glob_path(path: Union[str, list[str]], io_config: Optional[IOConfig] = None) -> DataFrame:
    """Creates a DataFrame of file paths and other metadata from a glob path.

    This method supports wildcards:

    1. `*` matches any number of any characters including none
    2. `?` matches any single character
    3. `[...]` matches any single character in the brackets
    4. `**` recursively matches any number of layers of directories

    The returned DataFrame will have the following columns:

    1. path: the path to the file/directory
    2. size: size of the object in bytes
    3. rows: the total rows of parquet object, it's None for other formats.

    Args:
        path (str|list): Path to files on disk (allows wildcards).
        io_config (IOConfig): Configuration to use when running IO with remote services

    Returns:
        DataFrame: DataFrame containing the path to each file as a row, along with other metadata parsed from the provided filesystem.

    Raises:
        FileNotFoundError: If none of files found with the glob paths.

    Examples:
        >>> df = daft.from_glob_path("/path/to/files/*.jpeg")
        >>> df = daft.from_glob_path("/path/to/files/**/*.jpeg")
        >>> df = daft.from_glob_path("/path/to/files/**/image-?.jpeg")
        >>> df = daft.from_glob_path(["/path/to/files/*.jpeg", "/path/to/others/*.jpeg"])
    """
    if isinstance(path, str):
        path = [path]

    context = get_context()
    io_config = context.daft_planning_config.default_io_config if io_config is None else io_config
    runner_io = context.get_or_create_runner().runner_io()
    file_infos = runner_io.glob_paths_details(path, io_config=io_config)
    file_infos_table = MicroPartition._from_pyrecordbatch(_PyRecordBatch.from_file_infos(file_infos))
    partition = LocalPartitionSet()
    partition.set_partition_from_table(0, file_infos_table)
    cache_entry = context.get_or_create_runner().put_partition_set_into_cache(partition)
    size_bytes = partition.size_bytes()
    num_rows = len(partition)

    assert size_bytes is not None, "In-memory data should always have non-None size in bytes"
    builder = LogicalPlanBuilder.from_in_memory_scan(
        cache_entry,
        schema=file_infos_table.schema(),
        num_partitions=partition.num_partitions(),
        size_bytes=size_bytes,
        num_rows=num_rows,
    )
    return DataFrame(builder)



from __future__ import annotations

import pytest

from daft.expressions import col
from tests.expressions.typing.conftest import assert_typing_resolve_vs_runtime_behavior


@pytest.mark.parametrize(
    "op",
    [
        pytest.param(lambda x: x.dt.day(), id="day"),
        pytest.param(lambda x: x.dt.month(), id="month"),
        pytest.param(lambda x: x.dt.year(), id="year"),
        pytest.param(lambda x: x.dt.day_of_week(), id="day_of_week"),
        pytest.param(lambda x: x.dt.date(), id="date"),
    ],
)
def test_dt_extraction_ops(unary_data_fixture, op):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=op(col(arg.name())),
        run_kernel=lambda: op(arg),
        resolvable=arg.datatype().is_temporal(),
    )



from __future__ import annotations

import contextlib
import csv
import decimal
import json
import os
import tempfile
import uuid

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as papq
import pytest

import daft
from daft.api_annotations import APITypeError
from daft.dataframe import DataFrame
from daft.datatype import DataType
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import UuidType

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


class MyObj:
    pass


class MyObj2:
    pass


class MyObjWithValue:
    def __init__(self, value: int):
        self.value = value

    def __eq__(self, other: MyObjWithValue):
        return isinstance(other, MyObjWithValue) and self.value == other.value


COL_NAMES = [
    "sepal_length",
    "sepal_width",
    "petal_length",
    "petal_width",
    "variety",
]


@pytest.mark.parametrize("read_method", ["read_csv", "read_json", "read_parquet"])
def test_load_missing(read_method):
    """Loading data from a missing filepath."""
    with pytest.raises(FileNotFoundError):
        getattr(daft, read_method)(str(uuid.uuid4()))


@pytest.mark.parametrize("data", [{"foo": [1, 2, 3]}, [{"foo": i} for i in range(3)], "foo"])
def test_error_thrown_create_dataframe_constructor(data) -> None:
    with pytest.raises(ValueError):
        DataFrame(data)


def test_wrong_input_type():
    with pytest.raises(APITypeError):
        daft.from_pydict("invalid input")


def test_create_dataframe_empty_list() -> None:
    with pytest.raises(ValueError):
        daft.read_parquet([])
    with pytest.raises(ValueError):
        daft.read_csv([])
    with pytest.raises(ValueError):
        daft.read_json([])


###
# List tests
###


def test_create_dataframe_list(valid_data: list[dict[str, float]]) -> None:
    df = daft.from_pylist(valid_data)
    assert len(df) == len(valid_data)
    assert len(df._preview.partition) == len(valid_data)
    assert set(df.column_names) == set(COL_NAMES)


def test_create_dataframe_list_data_longer_than_preview(valid_data: list[dict[str, float]]) -> None:
    valid_data = valid_data * 3
    df = daft.from_pylist(valid_data)
    assert len(df) == len(valid_data)
    assert len(df._preview.partition) == 8
    assert set(df.column_names) == set(COL_NAMES)


def test_create_dataframe_list_empty() -> None:
    df = daft.from_pylist([])
    assert df.column_names == []


def test_create_dataframe_list_ragged_keys() -> None:
    df = daft.from_pylist(
        [
            {"foo": 1},
            {"foo": 2, "bar": 1},
            {"foo": 3, "bar": 2, "baz": 1},
        ]
    )
    assert df.to_pydict() == {
        "foo": [1, 2, 3],
        "bar": [None, 1, 2],
        "baz": [None, None, 1],
    }


def test_create_dataframe_list_empty_dicts() -> None:
    df = daft.from_pylist([{}, {}, {}])
    assert df.column_names == []


def test_create_dataframe_list_non_dicts() -> None:
    with pytest.raises(ValueError) as e:
        daft.from_pylist([1, 2, 3])
    assert "Expected list of dictionaries of {column_name: value}" in str(e.value)


###
# Dict tests
###


def test_create_dataframe_pydict(valid_data: list[dict[str, float]]) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    df = daft.from_pydict(pydict)
    assert len(df) == len(valid_data)
    assert len(df._preview.partition) == len(valid_data)
    assert set(df.column_names) == set(COL_NAMES)


def test_create_dataframe_pydict_data_longer_than_preview(valid_data: list[dict[str, float]]) -> None:
    valid_data = valid_data * 3
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    df = daft.from_pydict(pydict)
    assert len(df) == len(valid_data)
    assert len(df._preview.partition) == 8
    assert set(df.column_names) == set(COL_NAMES)


def test_create_dataframe_empty_pydict() -> None:
    df = daft.from_pydict({})
    assert df.column_names == []


def test_create_dataframe_pydict_ragged_col_lens() -> None:
    with pytest.raises(ValueError) as e:
        daft.from_pydict({"foo": [1, 2], "bar": [1, 2, 3]})
    assert "Expected all columns to be of the same length" in str(e.value)


def test_create_dataframe_pydict_bad_columns() -> None:
    with pytest.raises(ValueError) as e:
        daft.from_pydict({"foo": "somestring"})
    assert "Creating a Series from data of type" in str(e.value)


###
# Arrow tests
###


@pytest.mark.parametrize("multiple", [False, True])
def test_create_dataframe_arrow(valid_data: list[dict[str, float]], multiple) -> None:
    t = pa.Table.from_pydict({k: [valid_data[i][k] for i in range(len(valid_data))] for k in valid_data[0].keys()})
    if multiple:
        t = [t, t, t]
    df = daft.from_arrow(t)
    if multiple:
        t = pa.concat_tables(t)
    assert len(df) == len(t)
    assert len(df._preview.partition) == (8 if multiple else len(t))
    assert set(df.column_names) == set(t.column_names)
    casted_field = t.schema.field("variety").with_type(pa.large_string())
    expected = t.cast(t.schema.set(t.schema.get_field_index("variety"), casted_field))
    # Check roundtrip.
    assert df.to_arrow() == expected


@pytest.mark.skipif(
    not pyarrow_supports_fixed_shape_tensor(),
    reason=f"Arrow version {ARROW_VERSION} doesn't support the canonical tensor extension type.",
)
def test_create_dataframe_arrow_tensor_canonical(valid_data: list[dict[str, float]]) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    shape = (2, 2)
    dtype = pa.fixed_shape_tensor(pa.int64(), shape)
    storage = pa.array([list(range(4 * i, 4 * (i + 1))) for i in range(len(valid_data))], pa.list_(pa.int64(), 4))
    fst = pa.ExtensionArray.from_storage(dtype, storage)
    pydict["tensor"] = fst
    t = pa.Table.from_pydict(pydict)
    df = daft.from_arrow(t)
    assert set(df.column_names) == set(t.column_names)
    assert df.schema()["tensor"].dtype == DataType.tensor(DataType.int64(), shape)
    casted_field = t.schema.field("variety").with_type(pa.large_string())
    expected = t.cast(t.schema.set(t.schema.get_field_index("variety"), casted_field))
    # Check roundtrip.
    assert df.to_arrow() == expected


def test_create_dataframe_arrow_extension_type(valid_data: list[dict[str, float]], uuid_ext_type: UuidType) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    storage = pa.array([f"{i}".encode() for i in range(len(valid_data))])
    pydict["obj"] = pa.ExtensionArray.from_storage(uuid_ext_type, storage)
    t = pa.Table.from_pydict(pydict)
    df = daft.from_arrow(t)
    assert set(df.column_names) == set(t.column_names)
    assert df.schema()["obj"].dtype == DataType.extension(
        uuid_ext_type.NAME, DataType.from_arrow_type(uuid_ext_type.storage_type), ""
    )
    casted_field = t.schema.field("variety").with_type(pa.large_string())
    expected = t.cast(t.schema.set(t.schema.get_field_index("variety"), casted_field))
    # Check roundtrip.
    assert df.to_arrow() == expected


class PyExtType(pa.PyExtensionType):
    def __init__(self):
        pa.PyExtensionType.__init__(self, pa.binary())

    def __reduce__(self):
        return PyExtType, ()


def test_create_dataframe_arrow_py_ext_type_raises(valid_data: list[dict[str, float]]) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    uuid_type = PyExtType()
    storage_array = pa.array([f"foo-{i}".encode() for i in range(len(valid_data))], pa.binary())
    arr = pa.ExtensionArray.from_storage(uuid_type, storage_array)
    pydict["obj"] = arr
    t = pa.Table.from_pydict(pydict)
    with pytest.raises(ValueError):
        daft.from_arrow(t)


def test_create_dataframe_arrow_unsupported_dtype(valid_data: list[dict[str, float]]) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    pydict["obj"] = [
        decimal.Decimal("12456789012345678901234567890123456789012345678901234567890") for _ in range(len(valid_data))
    ]
    t = pa.Table.from_pydict(pydict)
    df = daft.from_arrow(t)
    assert set(df.column_names) == set(t.column_names)
    # Type not natively supported, so should have Python object dtype.
    assert df.schema()["obj"].dtype == DataType.python()

    # Assert that it raises an error when trying to convert back to arrow
    with pytest.raises(ValueError):
        df.to_arrow()


###
# Pandas tests
###


@pytest.mark.parametrize("multiple", [False, True])
def test_create_dataframe_pandas(valid_data: list[dict[str, float]], multiple) -> None:
    pd_df = pd.DataFrame(valid_data)
    if multiple:
        pd_df = [pd_df, pd_df, pd_df]
    df = daft.from_pandas(pd_df)
    if multiple:
        pd_df = pd.concat(pd_df).reset_index(drop=True)
    assert len(df) == len(pd_df)
    assert len(df._preview.partition) == (8 if multiple else len(pd_df))
    assert set(df.column_names) == set(pd_df.columns)
    # Check roundtrip.
    pd.testing.assert_frame_equal(df.to_pandas(), pd_df)


def test_create_dataframe_pandas_py_object(valid_data: list[dict[str, float]]) -> None:
    pydict = {k: [item[k] for item in valid_data] for k in valid_data[0].keys()}
    pydict["obj"] = [MyObjWithValue(i) for i in range(len(valid_data))]
    pd_df = pd.DataFrame(pydict)
    df = daft.from_pandas(pd_df)
    # Type not natively supported, so should have Python object dtype.
    assert df.schema()["obj"].dtype == DataType.python()
    assert set(df.column_names) == set(pd_df.columns)
    # Check roundtrip.
    pd.testing.assert_frame_equal(df.to_pandas(), pd_df)


@pytest.mark.parametrize(
    ["data", "expected_dtype"],
    [
        pytest.param([None, 2, 3], DataType.int64(), id="arrow_int"),
        pytest.param([None, 1.0, 3.0], DataType.float64(), id="arrow_float"),
        pytest.param([None, 1.0, 3], DataType.float64(), id="arrow_mixed_numbers"),
        pytest.param([None, "b", "c"], DataType.string(), id="arrow_str"),
        pytest.param([None, None, None], DataType.null(), id="arrow_nulls"),
        pytest.param(np.array([1, 2, 3], dtype=np.int64), DataType.int64(), id="np_int"),
        pytest.param(np.array([None, "foo", "bar"], dtype=np.object_), DataType.string(), id="np_string"),
        pytest.param(pa.array([1, 2, 3]), DataType.int64(), id="pa_int"),
        pytest.param(pa.chunked_array([pa.array([1, 2, 3])]), DataType.int64(), id="pa_int_chunked"),
        pytest.param([None, MyObj(), MyObj()], DataType.python(), id="py_objs"),
        pytest.param([None, MyObj(), MyObj2()], DataType.python(), id="heterogenous_py_objs"),
        pytest.param(np.array([None, MyObj(), MyObj()], dtype=np.object_), DataType.python(), id="np_object"),
        pytest.param(
            [None, {"foo": 1}, {"bar": 1}],
            DataType.struct({"bar": DataType.int64(), "foo": DataType.int64()}),
            id="arrow_struct",
        ),
        pytest.param(
            [np.array([1], dtype=np.int64), np.array([2], dtype=np.int64), np.array([3], dtype=np.int64)],
            DataType.list(DataType.int64()),
            id="numpy_1d_arrays",
        ),
        pytest.param(pa.array([[1, 2, 3], [1, 2], [1]]), DataType.list(DataType.int64()), id="pa_nested"),
        pytest.param(
            pa.array([[("a", 1), ("b", 2)], [("c", 3), ("d", 4)]], type=pa.map_(pa.string(), pa.int64())),
            DataType.map(DataType.string(), DataType.int64()),
            id="pa_map",
        ),
        pytest.param(
            pa.array(
                [[("a", [("b", 1)]), ("c", [("d", 2)])], [("e", [("f", 3)]), ("g", [("h", 4)])]],
                type=pa.map_(pa.string(), pa.map_(pa.string(), pa.int64())),
            ),
            DataType.map(DataType.string(), DataType.map(DataType.string(), DataType.int64())),
            id="pa_nested_map",
        ),
        pytest.param(
            pa.chunked_array([pa.array([[1, 2, 3], [1, 2], [1]])]),
            DataType.list(DataType.int64()),
            id="pa_nested_chunked",
        ),
        pytest.param(np.ones((3, 3)), DataType.list(DataType.float64()), id="np_nested_1d"),
        pytest.param(np.ones((3, 3, 3)), DataType.tensor(DataType.float64()), id="np_nested_nd"),
    ],
)
def test_load_pydict_types(data, expected_dtype):
    data_dict = {"x": data}
    daft_df = daft.from_pydict(data_dict)
    daft_df.collect()
    collected_data = daft_df.to_pydict()

    assert list(collected_data.keys()) == ["x"]
    assert daft_df.schema()["x"].dtype == expected_dtype


###
# CSV tests
###
@contextlib.contextmanager
def create_temp_filename() -> str:
    with tempfile.TemporaryDirectory() as dir:
        yield os.path.join(dir, "tempfile")


def test_create_dataframe_csv(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(fname)
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_multiple_csvs(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                header = list(valid_data[0].keys())
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows([[item[col] for col in header] for item in valid_data])
                f.flush()

        df = daft.read_csv([f1name, f2name])
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == (len(valid_data) * 2)


def test_create_dataframe_csv_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(fname, file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert all(pd_df["file_path"] == fname)


def test_create_dataframe_multiple_csvs_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                header = list(valid_data[0].keys())
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows([[item[col] for col in header] for item in valid_data])
                f.flush()

        df = daft.read_csv([f1name, f2name], file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == (len(valid_data) * 2)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data) + [f2name] * len(valid_data)


def test_create_dataframe_csv_with_file_path_column_and_pushdowns(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                header = list(valid_data[0].keys())
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows([[item[col] for col in header] for item in valid_data])
                f.flush()

        df = daft.read_csv([f1name, f2name], file_path_column="file_path").where(daft.col("file_path") == f1name)
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data)


def test_create_dataframe_csv_with_file_path_column_duplicate_field_names() -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            data = [{"path": 1, "data": "a"}, {"path": 2, "data": "b"}, {"path": 3, "data": "c"}]
            header = list(data[0].keys())
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in data])
            f.flush()

        with pytest.raises(ValueError):
            # The file_path_column name is the same as a column in the table, which is not allowed
            daft.read_json(fname, file_path_column="path")


def test_create_dataframe_csv_generate_headers(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        cnames = [f"column_{i}" for i in range(1, 6)]
        df = daft.read_csv(fname, has_headers=False)
        assert df.column_names == cnames

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == cnames
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_column_projection(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f)
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        col_subset = COL_NAMES[:3]

        df = daft.read_csv(fname)
        df = df.select(*col_subset)
        assert df.column_names == col_subset

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == col_subset
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_custom_delimiter(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(fname, delimiter="\t")
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_provided_schema(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(
            fname,
            delimiter="\t",
            infer_schema=False,
            schema={
                "s_length": DataType.float32(),
                "s_width": DataType.string(),
                "p_length": DataType.float32(),
                "p_width": DataType.string(),
                "variety": DataType.string(),
            },
        )
        assert df.column_names == ["s_length", "s_width", "p_length", "p_width", "variety"]
        assert df.schema()["s_length"].dtype == DataType.float32()
        assert df.schema()["s_width"].dtype == DataType.string()
        assert df.schema()["p_length"].dtype == DataType.float32()
        assert df.schema()["p_width"].dtype == DataType.string()
        assert df.schema()["variety"].dtype == DataType.string()

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == ["s_length", "s_width", "p_length", "p_width", "variety"]
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_provided_schema_no_headers(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        schema_for_csv_without_headers = {
            "c_1": DataType.float64(),
            "c_2": DataType.float64(),
            "c_3": DataType.float64(),
            "c_4": DataType.float64(),
            "c_5": DataType.string(),
        }

        df = daft.read_csv(
            fname,
            delimiter="\t",
            infer_schema=False,
            schema=schema_for_csv_without_headers,
            has_headers=False,
        )
        assert df.column_names == list(schema_for_csv_without_headers.keys())

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == list(schema_for_csv_without_headers.keys())
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_schema_hints_partial(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(
            fname,
            delimiter="\t",
            infer_schema=True,
            schema={
                "sepal_length": DataType.float64(),
                "sepal_width": DataType.float64(),
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_schema_hints_override_types(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(
            fname,
            delimiter="\t",
            infer_schema=True,
            schema={
                "sepal_length": DataType.string(),  # Override the inferred float64 type to string
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)

        assert pd_df["sepal_length"].dtype == "object"
        assert pd_df["sepal_length"][0] == str(valid_data[0]["sepal_length"])


def test_create_dataframe_csv_schema_hints_ignore_random_hint(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        df = daft.read_csv(
            fname,
            delimiter="\t",
            infer_schema=True,
            schema={
                "foo": DataType.string(),  # Random column name that is not in the table
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_csv_without_schema_or_inference(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            header = list(valid_data[0].keys())
            writer = csv.writer(f, delimiter="\t")
            writer.writerow(header)
            writer.writerows([[item[col] for col in header] for item in valid_data])
            f.flush()

        with pytest.raises(ValueError):
            daft.read_csv(
                fname,
                delimiter="\t",
                infer_schema=False,
            )


###
# JSON tests
###


def test_create_dataframe_json(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(fname)
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_multiple_jsons(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                for data in valid_data:
                    f.write(json.dumps(data))
                    f.write("\n")
                f.flush()

        df = daft.read_json([f1name, f2name])
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == (len(valid_data) * 2)


def test_create_dataframe_json_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(fname, file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert all(pd_df["file_path"] == fname)


def test_create_dataframe_multiple_jsons_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                for data in valid_data:
                    f.write(json.dumps(data))
                    f.write("\n")
                f.flush()

        df = daft.read_json([f1name, f2name], file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == (len(valid_data) * 2)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data) + [f2name] * len(valid_data)


def test_create_dataframe_json_with_file_path_column_and_pushdowns(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                for data in valid_data:
                    f.write(json.dumps(data))
                    f.write("\n")
                f.flush()

        df = daft.read_json([f1name, f2name], file_path_column="file_path").where(daft.col("file_path") == f1name)
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data)


def test_create_dataframe_json_with_file_path_column_duplicate_field_names() -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            data = {"path": [1, 2, 3], "data": [4, 5, 6]}
            for i in range(len(data["path"])):
                f.write(json.dumps({k: data[k][i] for k in data}))
                f.write("\n")
            f.flush()

        with pytest.raises(ValueError):
            # The file_path_column name is the same as a column in the table, which is not allowed
            daft.read_json(fname, file_path_column="path")


def test_create_dataframe_json_column_projection(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        col_subset = COL_NAMES[:3]

        df = daft.read_json(fname)
        df = df.select(*col_subset)
        assert df.column_names == col_subset

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == col_subset
        assert len(pd_df) == len(valid_data)


@pytest.mark.skip
def test_create_dataframe_json_https() -> None:
    df = daft.read_json(
        "https://github.com/Eventual-Inc/mnist-json/raw/master/mnist_handwritten_test.json.gz",
    )
    df.collect()
    assert set(df.column_names) == {"label", "image"}
    assert len(df) == 10000


def test_create_dataframe_json_provided_schema(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(
            fname,
            infer_schema=False,
            schema={
                "sepal_length": DataType.float32(),
                "sepal_width": DataType.float32(),
                "petal_length": DataType.float32(),
                "petal_width": DataType.float32(),
                "variety": DataType.string(),
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_json_partial_schema(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(
            fname,
            infer_schema=True,
            schema={
                "sepal_length": DataType.float64(),
                "sepal_width": DataType.float64(),
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_json_schema_override_types(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(
            fname,
            infer_schema=True,
            schema={
                "sepal_length": DataType.string(),  # Override the inferred float64 type to string
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)

        assert pd_df["sepal_length"].dtype == "object"
        assert pd_df["sepal_length"][0] == str(valid_data[0]["sepal_length"])


def test_create_dataframe_json_schema_hints_ignore_random_hint(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in valid_data:
                f.write(json.dumps(data))
                f.write("\n")
            f.flush()

        df = daft.read_json(
            fname,
            infer_schema=True,
            schema={
                "foo": DataType.string(),  # Random column name that is not in the table
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_json_schema_hints_two_files() -> None:
    with create_temp_filename() as fname, create_temp_filename() as fname2:
        with open(fname, "w") as f:
            f.write(json.dumps({"foo": {"bar": "baz"}}))
            f.write("\n")
            f.flush()

        with open(fname2, "w") as f:
            f.write(json.dumps({"foo": {"bar2": "baz2"}}))
            f.write("\n")
            f.flush()

        # Without schema hints, schema inference should not pick up bar2
        assert daft.read_json([fname, fname2]).schema()["foo"].dtype == DataType.struct({"bar": DataType.string()})

        # With schema hints, bar2 should be included
        df = daft.read_json(
            [fname, fname2],
            infer_schema=True,
            schema={"foo": DataType.struct({"bar": DataType.string(), "bar2": DataType.string()})},
        )
        assert df.schema()["foo"].dtype == DataType.struct({"bar": DataType.string(), "bar2": DataType.string()})

        # When dataframe is materialized, the schema hints should be enforced and bar2 should be included
        df = df.select(df["foo"].struct.get("bar2"))
        df = df.where(df["bar2"].not_null()).collect()

        assert len(df) == 1
        assert df.to_pydict()["bar2"][0] == "baz2"


@pytest.mark.parametrize(
    "input,expected",
    [
        pytest.param(['{"foo": {}}', '{"foo": {}}'], {"foo": [{}, {}]}, id="AllEmptyObjects"),
        pytest.param(
            ['{"foo": {}}', '{"foo": {"bar":"baz"}}'],
            {"foo": [{"bar": None}, {"bar": "baz"}]},
            id="OneEmptyObject",
        ),
        pytest.param(
            ['{"foo": {}}', '{"foo":null}'],
            {"foo": [{}, None]},
            id="EmptyObjectAndNulls",
        ),
        pytest.param(
            ['{"foo": {"bar":{"baz":{}}}}', '{"foo": {"bar":{"baz":{}}}}'],
            {"foo": [{"bar": {"baz": {}}}, {"bar": {"baz": {}}}]},
            id="AllEmptyNestedObjects",
        ),
        pytest.param(
            ['{"foo": {"bar":{"baz":{}}}}', '{"foo": {"bar":{"baz":{"foo2":"bar2"}}}}'],
            {"foo": [{"bar": {"baz": {"foo2": None}}}, {"bar": {"baz": {"foo2": "bar2"}}}]},
            id="OneEmptyNestedObject",
        ),
    ],
)
def test_create_dataframe_json_empty_objects(input, expected) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            for data in input:
                f.write(data)
                f.write("\n")
            f.flush()

        df = daft.read_json(fname)
        pydict = df.to_pydict()

        assert pydict == expected


###
# Parquet tests
###


def test_create_dataframe_parquet(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, f.name)
            f.flush()

        df = daft.read_parquet(fname)
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_parquet_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, f.name)
            f.flush()

        df = daft.read_parquet(fname, file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert all(pd_df["file_path"] == fname)


def test_create_dataframe_multiple_parquets_with_file_path_column(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
                papq.write_table(table, f.name)
                f.flush()

        df = daft.read_parquet([f1name, f2name], file_path_column="file_path")
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == (len(valid_data) * 2)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data) + [f2name] * len(valid_data)


def test_create_dataframe_parquet_with_file_path_column_and_pushdowns(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
                papq.write_table(table, f.name)
                f.flush()

        df = daft.read_parquet([f1name, f2name], file_path_column="file_path").where(daft.col("file_path") == f1name)
        assert df.column_names == COL_NAMES + ["file_path"]

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES + ["file_path"]
        assert len(pd_df) == len(valid_data)
        assert pd_df["file_path"].to_list() == [f1name] * len(valid_data)


def test_create_dataframe_parquet_with_file_path_column_duplicate_field_names() -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({"path": [1, 2, 3], "data": [4, 5, 6]})
            papq.write_table(table, f.name)
            f.flush()

        with pytest.raises(ValueError):
            # The file_path_column name is the same as a column in the table, which is not allowed
            daft.read_parquet(fname, file_path_column="path")


def test_create_dataframe_parquet_with_filter(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, f.name)
            f.flush()

        df = daft.read_parquet(fname)
        assert df.column_names == COL_NAMES

        df = df.where(daft.col("sepal_length") > 4.8)

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data) - 1


def test_create_dataframe_multiple_parquets(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as f1name, create_temp_filename() as f2name:
        with open(f1name, "w") as f1, open(f2name, "w") as f2:
            for f in (f1, f2):
                table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
                papq.write_table(table, f.name)
                f.flush()

        df = daft.read_parquet([f1name, f2name])
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == (len(valid_data) * 2)


def test_create_dataframe_parquet_column_projection(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, fname)
            f.flush()

        col_subset = COL_NAMES[:3]

        df = daft.read_parquet(fname)
        df = df.select(*col_subset)
        assert df.column_names == col_subset

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == col_subset
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_parquet_provided_schema(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, fname)
            f.flush()

        df = daft.read_parquet(
            fname,
            infer_schema=False,
            schema={
                "sepal_length": DataType.float64(),
                "sepal_width": DataType.float64(),
                "petal_length": DataType.float64(),
                "petal_width": DataType.float64(),
                "variety": DataType.string(),
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_parquet_partial_schema(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, fname)
            f.flush()

        df = daft.read_parquet(
            fname,
            infer_schema=True,
            schema={
                "sepal_length": DataType.float64(),
                "sepal_width": DataType.float64(),
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_parquet_schema_override_types(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, fname)
            f.flush()

        df = daft.read_parquet(
            fname,
            infer_schema=True,
            schema={
                "sepal_length": DataType.string(),  # Override the inferred float64 type to string
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)

        assert pd_df["sepal_length"].dtype == "object"
        assert pd_df["sepal_length"][0] == str(valid_data[0]["sepal_length"])


def test_create_dataframe_parquet_schema_hints_ignore_random_hint(valid_data: list[dict[str, float]]) -> None:
    with create_temp_filename() as fname:
        with open(fname, "w") as f:
            table = pa.Table.from_pydict({col: [d[col] for d in valid_data] for col in COL_NAMES})
            papq.write_table(table, fname)
            f.flush()

        df = daft.read_parquet(
            fname,
            infer_schema=True,
            schema={
                "foo": DataType.string(),  # Random column name that is not in the table
            },
        )
        assert df.column_names == COL_NAMES

        pd_df = df.to_pandas()
        assert list(pd_df.columns) == COL_NAMES
        assert len(pd_df) == len(valid_data)


def test_create_dataframe_parquet_mismatched_schemas_no_pushdown():
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    with create_temp_filename() as f1, create_temp_filename() as f2:
        papq.write_table(pa.Table.from_pydict({"x": [1, 2, 3, 4]}), f1)
        papq.write_table(pa.Table.from_pydict({"y": [1, 2, 3, 4]}), f2)

        df = daft.read_parquet([f1, f2])
        assert df.schema().column_names() == ["x"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, None, None, None, None]}


def test_create_dataframe_parquet_read_mismatched_schemas_with_pushdown():
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    with create_temp_filename() as f1, create_temp_filename() as f2:
        papq.write_table(
            pa.Table.from_pydict(
                {
                    "x": [1, 2, 3, 4],
                    "y": [1, 2, 3, 4],
                    # NOTE: Need a column z here because Daft doesn't do a pushdown otherwise.
                    "z": [1, 1, 1, 1],
                }
            ),
            f1,
        )
        papq.write_table(pa.Table.from_pydict({"x": [5, 6, 7, 8]}), f2)

        df = daft.read_parquet([f1, f2])
        df = df.select("x", "y")  # Applies column selection pushdown on each read
        assert df.schema().column_names() == ["x", "y"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, 5, 6, 7, 8], "y": [1, 2, 3, 4, None, None, None, None]}


def test_create_dataframe_parquet_read_mismatched_schemas_with_pushdown_no_rows_read():
    # When we read files, we infer schema from the first file
    # Then when we read subsequent files, we want to be able to read the data still but add nulls for columns
    # that don't exist
    with create_temp_filename() as f1, create_temp_filename() as f2:
        papq.write_table(
            pa.Table.from_pydict(
                {
                    "x": [1, 2, 3, 4],
                    # NOTE: Need a column z here because Daft doesn't do a pushdown otherwise.
                    "z": [1, 1, 1, 1],
                }
            ),
            f1,
        )
        papq.write_table(pa.Table.from_pydict({"y": [1, 2, 3, 4]}), f2)

        df = daft.read_parquet([f1, f2])
        df = df.select("x")  # Applies column selection pushdown on each read
        assert df.schema().column_names() == ["x"]
        assert df.to_pydict() == {"x": [1, 2, 3, 4, None, None, None, None]}



import os

import griffe
import mkdocs_gen_files
from jinja2 import Environment, FileSystemLoader

# map of functions submodule to name displayed on docs
# also dictates category order
CATEGORY_TITLES = {"misc": "Miscellaneous", "columnar": "Columnar", "window": "Window", "llm": "LLM"}

# Set up Jinja2 environment
template_dir = os.path.join(os.path.dirname(__file__), "templates")
env = Environment(loader=FileSystemLoader(template_dir))
module = griffe.load("daft.functions")


def format_function_signature(fn: griffe.Function) -> str:
    """Format function signature with optional parameters in brackets."""
    result_parts = []
    in_optional_group = False
    optional_group = []

    for param in fn.parameters:
        # Check if parameter has a default value (is optional)
        is_optional = param.default is not None and param.kind.name not in ["var_positional", "var_keyword"]
        is_var = param.kind.name in ["var_positional", "var_keyword"]

        if is_optional:
            if not in_optional_group:
                in_optional_group = True
                optional_group = [param.name]
            else:
                optional_group.append(param.name)
        else:
            # Close any open optional group
            if in_optional_group:
                result_parts.append("[, " + ", ".join(optional_group) + "]")
                in_optional_group = False
                optional_group = []

            # Add comma if needed
            if result_parts:
                result_parts.append(", ")

            # Add the parameter
            if is_var:
                if param.kind.name == "var_positional":
                    result_parts.append("*" + param.name)
                else:  # var_keyword
                    result_parts.append("**" + param.name)
            else:
                result_parts.append(param.name)

    # Close any remaining optional group
    if in_optional_group:
        result_parts.append("[, " + ", ".join(optional_group) + "]")

    return "".join(result_parts)


def gen_index():
    categories = {name: [] for name in CATEGORY_TITLES.keys()}

    for fn_name in module.exports:
        fn: griffe.Function = module[fn_name]
        if not fn.is_function:
            raise ValueError(f"Expected all `daft.functions` exports to be functions, found: {fn_name}")

        category = fn.module.name
        if category not in categories:
            raise ValueError(
                f"`daft.functions.{category}` not in category titles mapping. Add it in `docs/gen_pages/gen_function_pages.py`"
            )

        # Create enhanced function object with formatted signature
        fn_data = {
            "name": fn.name,
            "signature": format_function_signature(fn),
            "description": fn.docstring.value.split("\n")[0] if fn.docstring else "",
        }
        categories[category].append(fn_data)

    empty_categories = [k for k, v in categories.items() if len(v) == 0]
    if len(empty_categories) > 0:
        raise ValueError(
            f"These function submodules no longer exist: {empty_categories}. Remove them from the category titles mapping in `docs/gen_pages/gen_function_pages.py`"
        )

    template = env.get_template("functions.md.j2")
    content = template.render(categories=categories, category_titles=CATEGORY_TITLES)

    with mkdocs_gen_files.open("api/functions/index.md", "w") as f:
        f.write(content)


def gen_function_page(fn: griffe.Function):
    template = env.get_template("function_page.md.j2")
    content = template.render(fn=fn)

    with mkdocs_gen_files.open(f"api/functions/{fn.name}.md", "w") as f:
        f.write(content)


def gen_nav_summary():
    nav = mkdocs_gen_files.Nav()

    nav["Functions"] = "index.md"
    for fn_name in module.exports:
        nav[fn_name] = f"{fn_name}.md"

    with mkdocs_gen_files.open("api/functions/SUMMARY.md", "w") as f:
        f.writelines(nav.build_literate_nav())


def main():
    gen_nav_summary()
    gen_index()

    for fn_name in module.exports:
        gen_function_page(module[fn_name])


main()



from __future__ import annotations

from daft.expressions import Expression, list_


def columns_sum(*exprs: Expression | str) -> Expression:
    """Sum values across columns.

    Args:
        exprs: The columns to sum.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_sum
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("sum", columns_sum("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ sum   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 7     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 9     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_sum requires at least one expression")
    return list_(*exprs).list.sum().alias("columns_sum")


def columns_mean(*exprs: Expression | str) -> Expression:
    """Average values across columns. Akin to `columns_avg`.

    Args:
        exprs: The columns to average.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_mean
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("mean", columns_mean("a", "b"))
        >>> df.show()
        ╭───────┬───────┬─────────╮
        │ a     ┆ b     ┆ mean    │
        │ ---   ┆ ---   ┆ ---     │
        │ Int64 ┆ Int64 ┆ Float64 │
        ╞═══════╪═══════╪═════════╡
        │ 1     ┆ 4     ┆ 2.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 3.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 4.5     │
        ╰───────┴───────┴─────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_mean requires at least one expression")
    return list_(*exprs).list.mean().alias("columns_mean")


def columns_avg(*exprs: Expression | str) -> Expression:
    """Average values across columns. Akin to `columns_mean`.

    Args:
        exprs: The columns to average across.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_avg
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("avg", columns_avg("a", "b"))
        >>> df.show()
        ╭───────┬───────┬─────────╮
        │ a     ┆ b     ┆ avg     │
        │ ---   ┆ ---   ┆ ---     │
        │ Int64 ┆ Int64 ┆ Float64 │
        ╞═══════╪═══════╪═════════╡
        │ 1     ┆ 4     ┆ 2.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 3.5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 4.5     │
        ╰───────┴───────┴─────────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_avg requires at least one expression")
    return list_(*exprs).list.mean().alias("columns_avg")


def columns_min(*exprs: Expression | str) -> Expression:
    """Find the minimum value across columns.

    Args:
        exprs: The columns to find the minimum of.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_min
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("min", columns_min("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ min   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 1     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 2     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 3     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_min requires at least one expression")
    return list_(*exprs).list.min().alias("columns_min")


def columns_max(*exprs: Expression | str) -> Expression:
    """Find the maximum value across columns.

    Args:
        exprs: The columns to find the maximum of.

    Examples:
        >>> import daft
        >>> from daft.functions import columns_max
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("max", columns_max("a", "b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ max   │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 4     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 6     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    if not exprs:
        raise ValueError("columns_max requires at least one expression")
    return list_(*exprs).list.max().alias("columns_max")



from __future__ import annotations

import pytest
import ray

import daft


def test_range():
    df = daft.range(5)
    assert df.to_pydict() == {"id": [0, 1, 2, 3, 4]}


def test_range_with_start():
    df = daft.range(2, 5)
    assert df.to_pydict() == {"id": [2, 3, 4]}


def test_range_with_step():
    df = daft.range(2, 10, 2)
    assert df.to_pydict() == {"id": [2, 4, 6, 8]}


def test_range_with_step_kwargs():
    df = daft.range(2, 10, step=2)
    assert df.to_pydict() == {"id": [2, 4, 6, 8]}


def test_with_start_end_and_step_kwargs():
    df = daft.range(start=2, end=10, step=2)
    assert df.to_pydict() == {"id": [2, 4, 6, 8]}


def test_with_no_args_raises_error():
    with pytest.raises(TypeError):
        daft.range()


def test_range_called_multiple_times():
    df = daft.range(10)
    assert df.count_rows() == 10
    assert df.count_rows() == 10
    assert len(df.collect()) == 10
    assert len(df.collect()) == 10

    # test with op
    df_with_filter = df.filter(daft.col("id") >= 5)
    assert df_with_filter.count_rows() == 5
    assert df_with_filter.count_rows() == 5
    assert len(df_with_filter.collect()) == 5
    assert len(df_with_filter.collect()) == 5


def test_range_partitioning_even():
    df = daft.range(0, 10, 1, 2)
    assert df.num_partitions() == 2

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 2

    assert partitions[0].to_pydict() == {"id": [0, 1, 2, 3, 4]}
    assert partitions[1].to_pydict() == {"id": [5, 6, 7, 8, 9]}


def test_range_partitioning_uneven():
    df = daft.range(0, 10, 1, 3)
    assert df.num_partitions() == 3

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 3

    assert partitions[0].to_pydict() == {"id": [0, 1, 2, 3]}
    assert partitions[1].to_pydict() == {"id": [4, 5, 6]}
    assert partitions[2].to_pydict() == {"id": [7, 8, 9]}


def test_range_partitioning_with_step_even():
    df = daft.range(0, 24, 2, 4)
    assert df.num_partitions() == 4

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 4

    assert partitions[0].to_pydict() == {"id": [0, 2, 4]}
    assert partitions[1].to_pydict() == {"id": [6, 8, 10]}
    assert partitions[2].to_pydict() == {"id": [12, 14, 16]}
    assert partitions[3].to_pydict() == {"id": [18, 20, 22]}


def test_range_partitioning_with_step_uneven():
    df = daft.range(0, 15, 2, 3)
    assert df.num_partitions() == 3

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 3

    assert partitions[0].to_pydict() == {"id": [0, 2, 4]}
    assert partitions[1].to_pydict() == {"id": [6, 8, 10]}
    assert partitions[2].to_pydict() == {"id": [12, 14]}


def test_range_partitioning_with_negative_step_even():
    df = daft.range(10, -2, -2, 2)
    assert df.num_partitions() == 2

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 2

    assert partitions[0].to_pydict() == {"id": [10, 8, 6]}
    assert partitions[1].to_pydict() == {"id": [4, 2, 0]}


def test_range_partitioning_with_negative_step_uneven():
    df = daft.range(15, 0, -2, 3)
    assert df.num_partitions() == 3

    partitions = list(df.iter_partitions())
    partitions = ray.get(partitions) if isinstance(partitions[0], ray.ObjectRef) else partitions
    assert len(partitions) == 3

    assert partitions[0].to_pydict() == {"id": [15, 13, 11]}
    assert partitions[1].to_pydict() == {"id": [9, 7, 5]}
    assert partitions[2].to_pydict() == {"id": [3, 1]}


def test_range_negative_step_validation():
    with pytest.raises(
        ValueError,
        match="daft.range\\(\\) with negative step -2 requires start \\(5\\) to be greater than end \\(10\\)",
    ):
        daft.range(5, 10, -2)

    # Should raise error when start == end for negative step
    with pytest.raises(
        ValueError,
        match="daft.range\\(\\) with negative step -1 requires start \\(5\\) to be greater than end \\(5\\)",
    ):
        daft.range(5, 5, -1)


def test_range_positive_step_validation():
    # Should raise error when start >= end for positive step
    with pytest.raises(
        ValueError,
        match="daft.range\\(\\) with positive step 2 requires start \\(10\\) to be less than end \\(5\\)",
    ):
        daft.range(10, 5, 2)

    # Should raise error when start == end for positive step
    with pytest.raises(
        ValueError,
        match="daft.range\\(\\) with positive step 1 requires start \\(5\\) to be less than end \\(5\\)",
    ):
        daft.range(5, 5, 1)


def test_range_zero_step_validation():
    with pytest.raises(
        ValueError,
        match="daft.range\\(\\) step parameter cannot be zero - use a positive or negative integer",
    ):
        daft.range(0, 10, 0)



from __future__ import annotations

import uuid

import pytest

from daft.recordbatch import MicroPartition

BUCKETS = ["retries-parquet-bucket"]


@pytest.mark.integration()
@pytest.mark.parametrize("status_code", [(400, "foo"), (403, "foo"), (404, "foo")])
@pytest.mark.parametrize("bucket", BUCKETS)
def test_non_retryable_errors(retry_server_s3_config, status_code: tuple[int, str], bucket: str):
    status_code, status_code_str = status_code
    data_path = f"s3://{bucket}/{status_code}/{status_code_str}/1/{uuid.uuid4()}"
    with pytest.raises((FileNotFoundError, ValueError)):
        MicroPartition.read_parquet(data_path, io_config=retry_server_s3_config)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "status_code",
    [
        # NOTE: Certain errors are only correctly retried when the correct str code is returned in the XML response
        # For these cases, we can pass them into our custom S3 endpoint so that it will echo it in the XML response on return
        (500, "InternalError"),
        (503, "SlowDown"),
        (504, "PLACEHOLDER_CODE - should not matter"),
        # TODO: [IO-RETRIES] We should also retry correctly on these error codes
        # These are marked as retryable error codes, see:
        # https://github.com/aws/aws-sdk-cpp/blob/8a9550f1db04b33b3606602ba181d68377f763df/src/aws-cpp-sdk-core/include/aws/core/http/HttpResponse.h#L113-L131
        # 509,
        # 408,
        # 429,
        # 419,
        # 440,
        # 598,
        # 599,
    ],
)
@pytest.mark.parametrize("bucket", BUCKETS)
def test_retryable_errors(retry_server_s3_config, status_code: tuple[int, str], bucket: str):
    # By default the SDK retries 3 times, so we should be able to tolerate NUM_ERRORS=2
    # Tweak this variable appropriately to match the retry policy
    NUM_ERRORS = 2
    status_code, status_code_str = status_code
    data_path = f"s3://{bucket}/{status_code}/{status_code_str}/{NUM_ERRORS}/{uuid.uuid4()}"

    MicroPartition.read_parquet(data_path, io_config=retry_server_s3_config)



from __future__ import annotations

from typing import Any, Literal

from daft.catalog import Catalog, Identifier, Table
from daft.context import get_context
from daft.daft import LogicalPlanBuilder as PyBuilder
from daft.daft import PySession, PyTableSource, sql_exec
from daft.dataframe import DataFrame
from daft.logical.builder import LogicalPlanBuilder
from daft.logical.schema import Schema
from daft.udf import UDF

__all__ = [
    "Session",
    "attach",
    "attach_catalog",
    "attach_function",
    "attach_table",
    "create_namespace",
    "create_namespace_if_not_exists",
    "create_table",
    "create_table_if_not_exists",
    "create_temp_table",
    "current_catalog",
    "current_namespace",
    "current_session",
    "detach_catalog",
    "detach_function",
    "detach_table",
    "drop_namespace",
    "drop_table",
    "get_catalog",
    "get_table",
    "has_catalog",
    "has_namespace",
    "has_table",
    "list_catalogs",
    "list_namespaces",
    "list_tables",
    "read_table",
    "set_catalog",
    "set_namespace",
    "set_session",
    "write_table",
]


class Session:
    """Session holds a connection's state and orchestrates execution of DataFrame and SQL queries against catalogs.

    Examples:
        >>> import daft
        >>> from daft.session import Session
        >>>
        >>> sess = Session()
        >>>
        >>> # Create a temporary table from a DataFrame
        >>> sess.create_temp_table("T", daft.from_pydict({"x": [1, 2, 3]}))
        >>>
        >>> # Read the table as a DataFrame
        >>> df = sess.read_table("T")
        >>>
        >>> # Execute an SQL query
        >>> sess.sql("SELECT * FROM T").show()
        >>>
        >>> # You can also retrieve the current session without creating a new one:
        >>> from daft.session import current_session
        >>> sess = current_session()
    """

    _session: PySession

    def __init__(self) -> None:
        self._session = PySession.empty()

    ###
    # factory methods
    ###

    @staticmethod
    def _from_pysession(session: PySession) -> Session:
        """Creates a session from a rust session wrapper."""
        s = Session.__new__(Session)
        s._session = session
        return s

    @staticmethod
    def _from_env() -> Session:
        """Creates a session from the environment's configuration."""
        # todo session builders, raise if DAFT_SESSION=0
        return Session()

    ###
    # exec
    ###

    def sql(self, sql: str) -> DataFrame | None:
        """Executes the SQL statement using this session.

        Args:
            sql (str): input SQL statement

        Returns:
            DataFrame: dataframe instance if this was a data statement (DQL, DDL, DML).
        """
        py_sess = self._session
        py_config = get_context().daft_planning_config
        py_object = sql_exec(sql, py_sess, {}, py_config)
        if py_object is None:
            return None
        elif isinstance(py_object, PyBuilder):
            return DataFrame(LogicalPlanBuilder(py_object))
        else:
            raise ValueError(f"Unsupported return type from sql exec: {type(py_object)}")

    ###
    # attach & detach
    ###

    def attach(self, object: Catalog | Table | UDF, alias: str | None = None) -> None:
        """Attaches a known attachable object like a Catalog, Table or UDF.

        Args:
            object (Catalog|Table|UDF): object which is attachable to a session

        Returns:
            None
        """
        if isinstance(object, Catalog):
            self.attach_catalog(object, alias)
        elif isinstance(object, Table):
            self.attach_table(object, alias)
        elif isinstance(object, UDF):
            self.attach_function(object, alias)
        else:
            raise ValueError(f"Cannot attach object with type {type(object)}")

    def attach_catalog(self, catalog: Catalog | object, alias: str | None = None) -> Catalog:
        """Attaches an external catalog to this session.

        Args:
            catalog (object): catalog instance or supported catalog object
            alias (str|None): optional alias for name resolution

        Returns:
            Catalog: new daft catalog instance
        """
        c = catalog if isinstance(catalog, Catalog) else Catalog._from_obj(catalog)
        a = alias if alias else c.name
        self._session.attach_catalog(c, a)
        return c

    def attach_table(self, table: Table | object, alias: str | None = None) -> Table:
        """Attaches an external table instance to this session.

        Args:
            table (Table | object): table instance or supported table object
            alias (str | None): optional alias for name resolution

        Returns:
            Table: new daft table instance
        """
        t = table if isinstance(table, Table) else Table._from_obj(table)
        a = alias if alias else t.name
        self._session.attach_table(t, a)
        return t

    def detach_catalog(self, alias: str) -> None:
        """Detaches the catalog from this session or raises if the catalog does not exist.

        Args:
            alias (str): catalog alias to detach
        """
        return self._session.detach_catalog(alias)

    def detach_table(self, alias: str) -> None:
        """Detaches the table from this session or raises if the table does not exist.

        Args:
            alias (str): catalog alias to detach
        """
        return self._session.detach_table(alias)

    ###
    # create_*
    ###

    def create_namespace(self, identifier: Identifier | str) -> None:
        """Creates a namespace in the current catalog."""
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot create a namespace without a current catalog")
        return catalog.create_namespace(identifier)

    def create_namespace_if_not_exists(self, identifier: Identifier | str) -> None:
        """Creates a namespace in the current catalog if it does not already exist."""
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot create a namespace without a current catalog")
        return catalog.create_namespace_if_not_exists(identifier)

    def create_table(self, identifier: Identifier | str, source: Schema | DataFrame, **properties: Any) -> Table:
        """Creates a table in the current catalog.

        If no namespace is specified, the current namespace is used.

        Returns:
            Table: the newly created table instance.
        """
        if not (catalog := self.current_catalog()):
            # TODO relax this constraint by joining with the catalog name
            raise ValueError("Cannot create a table without a current catalog")

        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        if len(identifier) == 1:
            if ns := self.current_namespace():
                identifier = ns + identifier

        return catalog.create_table(identifier, source, properties)

    def create_table_if_not_exists(
        self,
        identifier: Identifier | str,
        source: Schema | DataFrame,
        **properties: Any,
    ) -> Table:
        """Creates a table in the current catalog if it does not already exist.

        If no namespace is specified, the current namespace is used.

        Returns:
            Table: the newly created instance, or the existing table instance.
        """
        if not (catalog := self.current_catalog()):
            # TODO relax this constraint by joining with the catalog name
            raise ValueError("Cannot create a table without a current catalog")

        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        if len(identifier) == 1:
            if ns := self.current_namespace():
                identifier = ns + identifier

        return catalog.create_table_if_not_exists(identifier, source, properties)

    def create_temp_table(self, identifier: str, source: Schema | DataFrame) -> Table:
        """Creates a temp table scoped to this session's lifetime.

        Args:
            identifier (str): table identifier (name)
            source (TableSource|object): table source like a schema or dataframe

        Returns:
            Table: new table instance

        Examples:
            >>> import daft
            >>> from daft.session import Session
            >>> sess = Session()
            >>> sess.create_temp_table("T", daft.from_pydict({"x": [1, 2, 3]}))
            >>> sess.create_temp_table("S", daft.from_pydict({"y": [4, 5, 6]}))
            >>> sess.list_tables()
            [Identifier(''T''), Identifier(''S'')]

        Args:
            identifier (str): table identifier (name)
            source (Schema | DataFrame): table source is either a Schema or Dataframe

        Returns:
            Table: new table instance
        """
        if isinstance(source, Schema):
            py_source = PyTableSource.from_pyschema(source._schema)
        elif isinstance(source, DataFrame):
            py_source = PyTableSource.from_pybuilder(source._builder._builder)
        else:
            raise ValueError(
                f"Unsupported create_temp_table source, {type(source)}, expected either Schema or DataFrame."
            )
        return self._session.create_temp_table(identifier, py_source, replace=True)

    ###
    # drop_*
    ###

    def drop_namespace(self, identifier: Identifier | str) -> None:
        """Drop the given namespace in the current catalog.

        Args:
            identifier (Identifier|str): table identifier
        """
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot drop a namespace without a current catalog")
        return catalog.drop_namespace(identifier)

    def drop_table(self, identifier: Identifier | str) -> None:
        """Drop the given table in the current catalog.

        Args:
            identifier (Identifier|str): table identifier
        """
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot drop a table without a current catalog")
        # TODO join the identifier with the current namespace
        return catalog.drop_table(identifier)

    ###
    # session state
    ###

    def use(self, identifier: Identifier | str | None = None) -> None:
        """Use sets the current catalog and namespace."""
        if identifier is None:
            self.set_catalog(None)
            self.set_namespace(None)
            return
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)
        if len(identifier) == 1:
            self.set_catalog(str(identifier[0]))
        else:
            self.set_catalog(str(identifier[0]))
            self.set_namespace(identifier.drop(1))

    def current_catalog(self) -> Catalog | None:
        """Get the session's current catalog or None.

        Returns:
            Catalog: current catalog or None if one is not set
        """
        return self._session.current_catalog()

    def current_namespace(self) -> Identifier | None:
        """Get the session's current namespace or None.

        Returns:
            Identifier: current namespace or none if one is not set
        """
        ident = self._session.current_namespace()
        return Identifier._from_pyidentifier(ident) if ident else None

    ###
    # get_*
    ###

    def get_catalog(self, identifier: str) -> Catalog:
        """Returns the catalog or raises an exception if it does not exist.

        Args:
            identifier (str): catalog identifier (name)

        Returns:
            Catalog: The catalog object.

        Raises:
            ValueError: If the catalog does not exist.
        """
        return self._session.get_catalog(identifier)

    def get_table(self, identifier: Identifier | str) -> Table:
        """Returns the table or raises an exception if it does not exist.

        Args:
            identifier (Identifier|str): table identifier or identifier string

        Returns:
            Table: The table object.

        Raises:
            ValueError: If the table does not exist.
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)
        return self._session.get_table(identifier._ident)

    ###
    # has_*
    ###

    def has_catalog(self, identifier: str) -> bool:
        """Returns true if a catalog with the given identifier exists."""
        return self._session.has_catalog(identifier)

    def has_namespace(self, identifier: Identifier | str) -> bool:
        """Returns true if a namespace with the given identifier exists."""
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot call has_namespace without a current catalog")
        return catalog.has_namespace(identifier)

    def has_table(self, identifier: Identifier | str) -> bool:
        """Returns true if a table with the given identifier exists."""
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)
        return self._session.has_table(identifier._ident)

    ###
    # list_*
    ###

    def list_catalogs(self, pattern: str | None = None) -> list[str]:
        """Returns a list of available catalogs matching the pattern.

        This API currently returns a list of catalog names for backwards compatibility.
        In 0.5.0 this API will return a list of Catalog objects.

        Args:
            pattern (str): catalog name pattern

        Returns:
            list[str]: list of available catalog names
        """
        return self._session.list_catalogs(pattern)

    def list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        """Returns a list of matching namespaces in the current catalog."""
        if not (catalog := self.current_catalog()):
            raise ValueError("Cannot list namespaces without a current catalog")
        return catalog.list_namespaces(pattern)

    def list_tables(self, pattern: str | None = None) -> list[Identifier]:
        """Returns a list of available tables.

        Args:
            pattern (str): table name pattern

        Returns:
            list[Identifier]: list of available tables
        """
        return [Identifier._from_pyidentifier(i) for i in self._session.list_tables(pattern)]

    ###
    # read_*
    ###

    def read_table(self, identifier: Identifier | str, **options: Any) -> DataFrame:
        """Returns the table as a DataFrame or raises an exception if it does not exist.

        Args:
            identifier (Identifier|str): table identifier

        Returns:
            DataFrame:

        Raises:
            ValueError: If the tables does not exist.
        """
        return self.get_table(identifier).read(**options)

    ###
    # set_*
    ###

    def set_catalog(self, identifier: str | None) -> None:
        """Set the given catalog as current_catalog or raises an err if it does not exist.

        Args:
            identifier (str): sets the current catalog

        Raises:
            ValueError: If the catalog does not exist.
        """
        self._session.set_catalog(identifier)

    def set_namespace(self, identifier: Identifier | str | None) -> None:
        """Set the given namespace as current_namespace for table resolution.

        Args:
            identifier (Identifier | str): namespace identifier
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)
        self._session.set_namespace(identifier._ident if identifier else None)

    ###
    # write_*
    ###

    def write_table(
        self,
        identifier: Identifier | str,
        df: DataFrame,
        mode: Literal["append", "overwrite"] = "append",
        **options: dict[str, Any],
    ) -> None:
        """Writes the DataFrame to the table specified by the identifier.

        Args:
            identifier (Identifier|str): table identifier
            df (DataFrame): dataframe to write
            mode ("append"|"overwrite"): write mode, defaults to "append"
            **options (dict[str,Any]): additional, format-specific write options
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        self._session.get_table(identifier._ident).write(df, mode=mode, **options)

    ###
    # functions
    ###
    def attach_function(self, function: UDF, alias: str | None = None) -> None:
        """Attaches a Python function as a UDF in the current session."""
        self._session.attach_function(function, alias)

    def detach_function(self, alias: str) -> None:
        """Detaches a Python function as a UDF in the current session."""
        self._session.detach_function(alias)


###
# global active session
###

_SESSION: Session | None = None


def _session() -> Session:
    # Consider registering into the global context
    # ```
    # ctx = get_context()
    # if not ctx._session
    #     set_session(Session.from_env())
    # return ctx._session
    # ```
    global _SESSION
    if not _SESSION:
        _SESSION = Session._from_env()
    return _SESSION


###
# attach & detach
###


def attach(object: Catalog | Table | UDF, alias: str | None = None) -> None:
    """Attaches a known attachable object like a Catalog or Table."""
    return _session().attach(object, alias)


def attach_catalog(catalog: object | Catalog, alias: str | None = None) -> Catalog:
    """Attaches an external catalog to the current session."""
    return _session().attach_catalog(catalog, alias)


def attach_table(table: object | Table, alias: str | None = None) -> Table:
    """Attaches an external table to the current session."""
    return _session().attach_table(table, alias)


def detach_catalog(alias: str) -> None:
    """Detaches the catalog from the current session."""
    return _session().detach_catalog(alias)


def detach_table(alias: str) -> None:
    """Detaches the table from the current session."""
    return _session().detach_table(alias)


###
# create_*
###


def create_namespace(identifier: Identifier | str) -> None:
    """Creates a namespace in the current session's active catalog."""
    return _session().create_namespace(identifier)


def create_namespace_if_not_exists(identifier: Identifier | str) -> None:
    """Creates a namespace in the current session's active catalog if it does not already exist."""
    return _session().create_namespace_if_not_exists(identifier)


def create_table(identifier: Identifier | str, source: Schema | DataFrame, **properties: Any) -> Table:
    """Creates a table in the current session's active catalog and namespace."""
    return _session().create_table(identifier, source, **properties)


def create_table_if_not_exists(identifier: Identifier | str, source: Schema | DataFrame, **properties: Any) -> Table:
    """Creates a table in the current session's active catalog and namespace if it does not already exist."""
    return _session().create_table_if_not_exists(identifier, source, **properties)


def create_temp_table(identifier: str, source: Schema | DataFrame) -> Table:
    """Creates a temp table scoped to current session's lifetime."""
    return _session().create_temp_table(identifier, source)


###
# drop_*
###


def drop_namespace(identifier: Identifier | str) -> None:
    """Drops the namespace in the current session's active catalog."""
    return _session().drop_namespace(identifier)


def drop_table(identifier: Identifier | str) -> None:
    """Drops the table in the current session's active catalog."""
    return _session().drop_namespace(identifier)


###
# session state
###


def current_catalog() -> Catalog | None:
    """Returns the session's current catalog or None."""
    return _session().current_catalog()


def current_namespace() -> Identifier | None:
    """Returns the session's current namespace or None."""
    return _session().current_namespace()


def current_session() -> Session:
    """Returns the global context's current session."""
    return _session()


###
# get_*
###


def get_catalog(identifier: str) -> Catalog:
    """Returns the catalog from the current session or raises an exception if it does not exist."""
    return _session().get_catalog(identifier)


def get_table(identifier: Identifier | str) -> Table:
    """Returns the table from the current session or raises an exception if it does not exist."""
    return _session().get_table(identifier)


###
# has_*
###


def has_catalog(identifier: str) -> bool:
    """Returns true if a catalog with the given identifier exists in the current session."""
    return _session().has_catalog(identifier)


def has_namespace(identifier: Identifier | str) -> bool:
    """Returns true if a namespace with the given identifier exists in the current session."""
    return _session().has_namespace(identifier)


def has_table(identifier: Identifier | str) -> bool:
    """Returns true if a table with the given identifier exists in the current session."""
    return _session().has_table(identifier)


###
# list_*
###


def list_catalogs(pattern: str | None = None) -> list[str]:
    """Returns a list of available catalogs in the current session."""
    return _session().list_catalogs(pattern)


def list_namespaces(pattern: str | None = None) -> list[Identifier]:
    """Returns a list of matching namespaces in the current catalog."""
    return _session().list_namespaces(pattern)


def list_tables(pattern: str | None = None) -> list[Identifier]:
    """Returns a list of available tables in the current session."""
    return _session().list_tables(pattern)


###
# read_*
###


def read_table(identifier: Identifier | str, **options: Any) -> DataFrame:
    """Returns the table as a DataFrame or raises an exception if it does not exist."""
    return _session().read_table(identifier, **options)


###
# write_*
###


def write_table(
    identifier: Identifier | str,
    df: DataFrame,
    mode: Literal["append", "overwrite"] = "append",
    **options: Any,
) -> None:
    """Writes the DataFrame to the table specified with the identifier."""
    _session().write_table(identifier, df, mode, **options)


###
# set_*
###


def set_catalog(identifier: str | None) -> None:
    """Set the given catalog as current_catalog for the current session or raises an if it does not exist."""
    _session().set_catalog(identifier)


def set_namespace(identifier: Identifier | str | None) -> None:
    """Set the given namespace as current_namespace for the current session."""
    _session().set_namespace(identifier)


def set_session(session: Session) -> None:
    """Sets the global context's current session."""
    # Consider registering into the global context.
    # ```
    # ctx = get_context()
    # with ctx._lock:
    #     ctx._session = session
    # ```
    global _SESSION
    _SESSION = session


###
# functions
###


def attach_function(function: UDF, alias: str | None = None) -> None:
    """Attaches a Python function as a UDF in the current session."""
    _session().attach_function(function, alias)


def detach_function(alias: str) -> None:
    """Detaches a Python function as a UDF in the current session."""
    _session().detach_function(alias)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from traceback import TracebackException


class ExpressionTypeError(Exception):
    pass


class UDFException(Exception):
    """An Daft exception raised when a UDF raises an exception.

    Also provides additional context about the error. The original exception is either available as:
    - The original exception via `__cause__` if running in the same process
    - A replica & rendered traceback via `tb_info` if running in a different process
    Both are accessible via `original_exception`
    """

    def __init__(self, message: str, tb_info: TracebackException | None = None):
        super().__init__(message)
        self.message = message
        self.tb_info = tb_info

    @property
    def original_exception(self) -> BaseException | None:
        """The original exception that was raised by the UDF."""
        # We except every creation of UDFException to be `raise UDFException(...) from ...`
        return self.__cause__

    def __str__(self) -> str:
        if self.tb_info:
            return (
                "\n".join(self.tb_info.format())
                + "\nThe above exception was the direct cause of the following exception:\n"
                + f"\n{self.message}"
            )
        else:
            return self.message



from __future__ import annotations

import time

import pytest

import daft
from daft.expressions import col
from daft.internal.gpu import cuda_visible_devices
from tests.conftest import get_tests_daft_runner_name
from tests.ray.autoscaling_cluster import autoscaling_cluster_context

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() != "ray",
    reason="Autoscaling tests require Ray runner to be in use",
)


def test_basic_autoscaling_cluster():
    """Test basic AutoscalingCluster functionality."""
    head_resources = {"CPU": 0}
    worker_node_types = {
        "worker": {
            "resources": {"CPU": 1},
            "node_config": {},
            "min_workers": 0,
            "max_workers": 1,
        }
    }

    with autoscaling_cluster_context(head_resources, worker_node_types):
        # Test basic Daft operations on the autoscaling cluster
        df = daft.from_pydict({"x": [1, 2, 3, 4, 5], "y": [6, 7, 8, 9, 10]})
        result = df.filter(col("x") > 3).to_pydict()
        assert result["x"] == [4, 5]
        assert result["y"] == [9, 10]


def test_basic_autoscaling_cluster_with_existing_workers():
    head_resources = {"CPU": 0}
    worker_node_types = {
        "worker1": {
            "resources": {"CPU": 1},
            "node_config": {},
            "min_workers": 1,
            "max_workers": 4,
        }
    }

    with autoscaling_cluster_context(head_resources, worker_node_types):

        @daft.udf(return_dtype=daft.DataType.list(daft.DataType.string()))
        def fake_udf_needs_2_cpus(x):
            time.sleep(1)
            return x

        # Test basic Daft operations on the autoscaling cluster
        df = daft.from_pydict({"x": [i for i in range(100)]}).repartition(50, "x")
        df.select(fake_udf_needs_2_cpus(col("x"))).to_pydict()


def test_basic_autoscaling_gpu_cluster():
    """Test basic AutoscalingCluster GPU functionality."""
    head_resources = {"CPU": 0, "GPU": 0}
    worker_node_types = {
        "worker": {
            "resources": {"CPU": 1, "GPU": 2},
            "node_config": {},
            "min_workers": 0,
            "max_workers": 1,
        }
    }

    with autoscaling_cluster_context(head_resources, worker_node_types):
        # Test basic Daft operations on the autoscaling cluster
        @daft.udf(return_dtype=daft.DataType.list(daft.DataType.string()), num_gpus=2)
        def fake_gpu_udf(x):
            visible_devices = cuda_visible_devices()
            return [visible_devices] * len(x)

        df = daft.from_pydict({"x": [1, 2, 3, 4, 5], "y": [6, 7, 8, 9, 10]})
        result = df.select(fake_gpu_udf(col("x"))).to_pydict()
        assert result["x"] == [["0", "1"]] * 5



"""The `daft.catalog` module contains functionality for Catalogs.

A Catalog can be understood as a system/service for users to discover, access and query their data.
Most commonly, users' data is represented as a "table". Some more modern Catalogs such as Unity Catalog
also expose other types of data including files, ML models, registered functions and more.

Examples of Catalogs include AWS Glue, Hive Metastore, Apache Iceberg REST and Unity Catalog.

**Catalog**

```python
# without any qualifiers, the default catalog and namespace are used.
daft.read_table("my_table")

# with a qualified identifier (uses default catalog)
daft.read_table("my_namespace.my_table")

# with a fully qualified identifier
daft.read_table("my_catalog.my_namespace.my_table")
```

**Named Tables**

Daft allows also the registration of temporary tables, which have no catalog associated with them.

Note that temporary tables take precedence over catalog tables when resolving unqualified names.

```python
df = daft.from_pydict({"foo": [1, 2, 3]})

# Attach the DataFrame for use across other APIs.
daft.attach_table(df, "my_table")

# Your table is now accessible from Daft-SQL, or Daft's `read_table`
df1 = daft.read_table("my_table")
df2 = daft.sql("SELECT * FROM my_table")
```
"""

from __future__ import annotations

import warnings

from abc import ABC, abstractmethod
from collections.abc import Sequence
from daft.daft import PyIdentifier

from daft.dataframe import DataFrame

from typing import TYPE_CHECKING, Any, Literal, overload

from daft.logical.schema import Schema

if TYPE_CHECKING:
    from daft.utils import ColumnInputType
    from daft.convert import InputListType


__all__ = [
    "Catalog",
    "Identifier",
    "Schema",
    "Table",
]


Properties = dict[str, Any]


class NotFoundError(Exception):
    """Raised when some catalog object is not able to be found."""


class Catalog(ABC):
    """Interface for Python catalog implementations.

    A Catalog is a service for discovering, accessing, and querying
    tabular and non-tabular data. You can instantiate a Catalog using
    one of the static `from_` methods.

    Examples:
        >>> import daft
        >>> from daft.catalog import Catalog
        >>>
        >>> data = {"users": {"id": [1, 2, 3], "name": ["a", "b", "c"]}}
        >>> catalog = Catalog.from_pydict(data)
        >>> catalog.list_tables()
        ['users']
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns the catalog's name."""

    @abstractmethod
    def _create_namespace(self, ident: Identifier) -> None:
        """Create a namespace in the catalog, erroring if the namespace already exists."""

    @abstractmethod
    def _create_table(self, ident: Identifier, schema: Schema, properties: Properties | None = None) -> Table:
        """Create a table in the catalog, erroring if the table already exists."""

    @abstractmethod
    def _drop_namespace(self, ident: Identifier) -> None:
        """Remove a namespace from the catalog, erroring if the namespace did not exist."""

    @abstractmethod
    def _drop_table(self, ident: Identifier) -> None:
        """Remove a table from the catalog, erroring if the table did not exist."""

    @abstractmethod
    def _get_table(self, ident: Identifier) -> Table:
        """Get a table from the catalog."""

    @abstractmethod
    def _has_namespace(self, ident: Identifier) -> bool:
        """Check if a namespace exists in the catalog."""

    @abstractmethod
    def _has_table(self, ident: Identifier) -> bool:
        """Check if a table exists in the catalog."""

    @abstractmethod
    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        """List all namespaces in the catalog. When a pattern is specified, list only namespaces matching the pattern."""

    @abstractmethod
    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        """List all tables in the catalog. When a pattern is specified, list only tables matching the pattern."""

    @staticmethod
    def from_pydict(tables: dict[Identifier | str, object], name: str = "default") -> Catalog:
        """Returns an in-memory catalog from a dictionary of table-like objects.

        The table-like objects can be pydicts, dataframes, or a Table implementation.
        For qualified tables, namespaces are created if necessary.

        Args:
            tables (dict[str,object]): a dictionary of table-like objects (pydicts, dataframes, and tables)

        Returns:
            Catalog: new catalog instance with name 'default'

        Examples:
            >>> import daft
            >>> from daft.catalog import Catalog, Table
            >>>
            >>> dictionary = {"x": [1, 2, 3]}
            >>> dataframe = daft.from_pydict(dictionary)
            >>> table = Table.from_df("temp", dataframe)
            >>>
            >>> catalog = Catalog.from_pydict(
            ...     {
            ...         "R": dictionary,
            ...         "S": dataframe,
            ...         "T": table,
            ...     }
            ... )
            >>> catalog.list_tables()
            ['R', 'S', 'T']

        """
        from daft.catalog.__internal import MemoryCatalog

        catalog = MemoryCatalog._new(name)

        for ident, source in tables.items():
            ident = ident if isinstance(ident, Identifier) else Identifier.from_str(ident)

            # has namespace, create one if it doesn't exist
            if len(ident) > 1:
                namespace = Identifier(*ident[:-1])
                catalog.create_namespace_if_not_exists(namespace)

            df: DataFrame
            if isinstance(source, Table):
                df = source.read()
            elif isinstance(source, DataFrame):
                df = source
            elif isinstance(source, dict):
                df = DataFrame._from_pydict(source)
            else:
                raise ValueError(f"Unsupported table source {type(source)}")

            catalog.create_table(ident, df)

        return catalog

    @staticmethod
    def from_iceberg(catalog: object) -> Catalog:
        """Create a Daft Catalog from a PyIceberg catalog object.

        Args:
            catalog (object): a PyIceberg catalog instance

        Returns:
            Catalog: a new Catalog instance backed by the PyIceberg catalog.

        Examples:
            >>> from pyiceberg.catalog import load_catalog
            >>> iceberg_catalog = load_catalog("my_iceberg_catalog")
            >>> catalog = Catalog.from_iceberg(iceberg_catalog)
        """
        try:
            from daft.catalog.__iceberg import IcebergCatalog

            return IcebergCatalog._from_obj(catalog)
        except ImportError:
            raise ImportError("Iceberg support not installed: pip install -U 'daft[iceberg]'")

    @staticmethod
    def from_unity(catalog: object) -> Catalog:
        """Create a Daft Catalog from a Unity Catalog client.

        Args:
            catalog (object): a Unity Catalog client instance

        Returns:
            Catalog: a new Catalog instance backed by the Unity catalog.

        Examples:
            >>> from unity_sdk import UnityCatalogClient
            >>> unity_client = UnityCatalogClient(...)
            >>> catalog = Catalog.from_unity(unity_client)

        """
        try:
            from daft.catalog.__unity import UnityCatalog

            return UnityCatalog._from_obj(catalog)
        except ImportError:
            raise ImportError("Unity support not installed: pip install -U 'daft[unity]'")

    @staticmethod
    def from_s3tables(
        table_bucket_arn: str,
        client: object | None = None,
        session: object | None = None,
    ) -> Catalog:
        """Creates a Daft Catalog from S3 Tables bucket ARN, with optional client or session.

        If neither a boto3 client nor session is provided, the Iceberg REST
        client will be used under the hood.

        Args:
            table_bucket_arn (str): ARN of the S3 Tables bucket
            client (object, optional): a boto3 client
            session (object, optional): a boto3 session

        Returns:
            Catalog: a new Catalog instance backed by S3 Tables.

        Examples:
            >>> arn = "arn:aws:s3:::my-s3tables-bucket"
            >>> catalog = Catalog.from_s3tables(arn)
            >>> catalog.list_tables()
        """
        try:
            from daft.catalog.__s3tables import S3Catalog

            if client is not None and session is not None:
                raise ValueError("Can provide either a client or session but not both.")
            elif client is not None:
                return S3Catalog.from_client(table_bucket_arn, client)
            elif session is not None:
                return S3Catalog.from_session(table_bucket_arn, session)
            else:
                return S3Catalog.from_arn(table_bucket_arn)
        except ImportError:
            raise ImportError("S3 Tables support not installed: pip install -U 'daft[aws]'")

    @staticmethod
    def from_glue(
        name: str,
        client: object | None = None,
        session: object | None = None,
    ) -> Catalog:
        """Creates a Daft Catalog backed by the AWS Glue service, with optional client or session.

        Terms:
            - AWS Glue          -> Daft Catalog
            - AWS Glue Database -> Daft Namespace
            - AWS Glue Table    -> Daft Table

        Args:
            name (str): glue database name
            client: optional boto3 client
            session: optional boto3 session

        Returns:
            Catalog: new daft catalog instance backed by AWS Glue.
        """
        try:
            from daft.catalog.__glue import GlueCatalog

            if client is not None and session is not None:
                raise ValueError("Can provide either a client or session but not both.")
            elif client is not None:
                return GlueCatalog.from_client(name, client)
            elif session is not None:
                return GlueCatalog.from_session(name, session)
            else:
                raise ValueError("Must provide either a client or session.")
        except ImportError:
            raise ImportError("AWS Glue support not installed: pip install -U 'daft[aws]'")

    @staticmethod
    def _from_obj(obj: object) -> Catalog:
        """Returns a Daft Catalog from a supported object type or raises a ValueError."""
        for factory in (Catalog.from_iceberg, Catalog.from_unity):
            try:
                return factory(obj)
            except ValueError:
                pass
            except ImportError:
                pass
        raise ValueError(
            f"Unsupported catalog type: {type(obj)}; please ensure all required extra dependencies are installed."
        )

    ###
    # create_*
    ###

    def create_namespace(self, identifier: Identifier | str) -> None:
        """Creates a namespace in this catalog.

        Args:
            identifier (Identifier | str): namespace identifier
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        self._create_namespace(identifier)

    def create_namespace_if_not_exists(self, identifier: Identifier | str) -> None:
        """Creates a namespace in this catalog if it does not already exist.

        Args:
            identifier (Identifier | str): namespace identifier
        """
        if not self.has_namespace(identifier):
            self.create_namespace(identifier)

    def create_table(
        self,
        identifier: Identifier | str,
        source: Schema | DataFrame,
        properties: Properties | None = None,
    ) -> Table:
        """Creates a table in this catalog.

        Args:
            identifier (Identifier | str): table identifier
            source (Schema | DataFrame): table source object such as a Schema or DataFrame.

        Returns:
            Table: new table instance.
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)
        schema = source.schema() if isinstance(source, DataFrame) else source

        table = self._create_table(identifier, schema, properties)
        if isinstance(source, DataFrame):
            table.append(source)

        return table

    def create_table_if_not_exists(
        self,
        identifier: Identifier | str,
        source: Schema | DataFrame,
        properties: Properties | None = None,
    ) -> Table:
        """Creates a table in this catalog if it does not already exist.

        Args:
            identifier (Identifier | str): table identifier
            source (Schema | DataFrame): table source object such as a Schema or DataFrame.

        Returns:
            Table: the existing table (if exists) or the new table instance.
        """
        if self.has_table(identifier):
            return self.get_table(identifier)
        else:
            return self.create_table(identifier, source, properties)

    ###
    # has_*
    ###

    def has_namespace(self, identifier: Identifier | str) -> bool:
        """Returns True if the namespace exists, otherwise False."""
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        return self._has_namespace(identifier)

    def has_table(self, identifier: Identifier | str) -> bool:
        """Returns True if the table exists, otherwise False."""
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        return self._has_table(identifier)

    ###
    # drop_*
    ###

    def drop_namespace(self, identifier: Identifier | str) -> None:
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        self._drop_namespace(identifier)

    def drop_table(self, identifier: Identifier | str) -> None:
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        self._drop_table(identifier)

    ###
    # get_*
    ###

    def get_table(self, identifier: Identifier | str) -> Table:
        """Get a table by its identifier or raises if the table does not exist.

        Args:
            identifier (Identifier|str): table identifier

        Returns:
            Table: matched table or raises if the table does not exist.
        """
        if isinstance(identifier, str):
            identifier = Identifier.from_str(identifier)

        return self._get_table(identifier)

    ###
    # list_*
    ###

    def list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        """List namespaces in the catalog which match the given pattern.

        Args:
            pattern (str): pattern to match such as a namespace prefix

        Returns:
            list[Identifier]: list of namespace identifiers matching the pattern.
        """
        return self._list_namespaces(pattern)

    def list_tables(self, pattern: str | None = None) -> list[Identifier]:
        """List tables in the catalog which match the given pattern.

        Args:
            pattern (str): pattern to match such as a namespace prefix

        Returns:
            list[str]: list of table identifiers matching the pattern.
        """
        return self._list_tables(pattern)

    ###
    # read_*
    ###

    def read_table(self, identifier: Identifier | str, **options: dict[str, Any]) -> DataFrame:
        """Returns the table as a DataFrame or raises an exception if it does not exist."""
        return self.get_table(identifier).read(**options)

    ###
    # write_*
    ###

    def write_table(
        self,
        identifier: Identifier | str,
        df: DataFrame,
        mode: Literal["append", "overwrite"] = "append",
        **options: dict[str, Any],
    ) -> None:
        return self.get_table(identifier).write(df, mode=mode, **options)

    ###
    # python methods
    ###

    def __repr__(self) -> str:
        return f"Catalog('{self.name}')"


class Identifier(Sequence[str]):
    """A reference (path) to a catalog object.

    Examples:
        >>> id = Identifier("a", "b")
        >>> assert len(id) == 2
    """

    _ident: PyIdentifier

    def __init__(self, *parts: str):
        """Creates an Identifier from its parts.

        Examples:
            >>> from daft.catalog import Identifier
            >>> Identifier("namespace", "table")
        """
        if len(parts) < 1:
            raise ValueError("Identifier requires at least one part.")
        self._ident = PyIdentifier(parts)

    @staticmethod
    def _from_pyidentifier(ident: PyIdentifier) -> Identifier:
        i = Identifier.__new__(Identifier)
        i._ident = ident
        return i

    @staticmethod
    def from_sql(input: str, normalize: bool = False) -> Identifier:
        """Parses an Identifier from an SQL string, normalizing to lowercase if specified.

        Args:
            input (str): input sql string
            normalize (bool): flag to case-normalize the identifier text

        Returns:
            Identifier: new identifier instance

        Examples:
            >>> from daft.catalog import Identifier
            >>> Identifier.from_sql("namespace.table") == Identifier("namespace", "table")
            >>> Identifier.from_sql('"a.b"') == Identifier('"a.b."')
            >>> Identifier.from_sql('ABC."xYz"', normalize=True) == Identifier("abc", "xYz")

        """
        i = Identifier.__new__(Identifier)
        i._ident = PyIdentifier.from_sql(input, normalize)
        return i

    @staticmethod
    def from_str(input: str) -> Identifier:
        """Parses an Identifier from a dot-delimited Python string without normalization.

        Args:
            input (str): input identifier string

        Returns:
            Identifier: new identifier instance

        Examples:
            >>> from daft.catalog import Identifier
            >>> Identifier.from_str("namespace.table") == Identifier("namespace", "table")

        """
        return Identifier(*input.split("."))

    def drop(self, n: int = 1) -> Identifier:
        """Returns a new Identifier with the first n parts removed.

        Args:
            n (int): Number of parts to drop from the beginning. Defaults to 1.

        Returns:
            Identifier: A new Identifier with the first n parts removed.

        Raises:
            ValueError: If dropping n parts would result in an empty Identifier.
        """
        if n <= 0:
            return Identifier(*self)
        if n >= len(self):
            raise ValueError(f"Cannot drop {n} parts from Identifier with {len(self)} parts")
        parts = tuple(self)
        return Identifier(*parts[n:])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Identifier):
            return False
        return self._ident.eq(other._ident)

    @overload
    def __getitem__(self, index: int, /) -> str: ...

    @overload
    def __getitem__(self, index: slice, /) -> Sequence[str]: ...

    def __getitem__(self, index: int | slice, /) -> str | Sequence[str]:
        if isinstance(index, int):
            return self._ident.getitem(index)
        parts = tuple(self)
        return parts[index]

    def __len__(self) -> int:
        return self._ident.__len__()

    def __add__(self, suffix: Identifier) -> Identifier:
        return Identifier(*(tuple(self) + tuple(suffix)))

    def __repr__(self) -> str:
        return f"Identifier('{self._ident.__repr__()}')"

    def __str__(self) -> str:
        return ".".join(self)


class Table(ABC):
    """Interface for python table implementations."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns the table's name."""

    @abstractmethod
    def schema(self) -> Schema:
        """Returns the table's schema."""

    @staticmethod
    def from_pydict(name: str, data: dict[str, InputListType]) -> Table:
        """Returns a read-only table backed by the given data.

        Args:
            name (str): table table
            data dict[str,object]: keys are column names and the values are python lists, numpy arrays, or arrow arrays.

        Returns:
            DataFrame: new read-only table instance

        Examples:
            >>> from daft.catalog import Table
            >>> table = Table.from_pydict({"foo": [1, 2]})
            >>> table.show()
            ╭───────╮
            │ foo   │
            │ ---   │
            │ Int64 │
            ╞═══════╡
            │ 1     │
            ├╌╌╌╌╌╌╌┤
            │ 2     │
            ╰───────╯
            <BLANKLINE>
            (Showing first 2 of 2 rows)

        """
        df = DataFrame._from_pydict(data)
        return Table.from_df(name, df)

    @staticmethod
    def from_df(name: str, dataframe: DataFrame) -> Table:
        """Returns a read-only table backed by the DataFrame.

        Args:
            name (str): table name
            dataframe (DataFrame): table source dataframe

        Returns:
            Table: new table instance

        Examples:
            >>> import daft
            >>> from daft.catalog import Table
            >>> Table.from_df("my_table", daft.from_pydict({"x": [1, 2, 3]}))

        """
        from daft.catalog.__internal import MemoryTable

        table = MemoryTable._new(name, dataframe.schema())
        table.append(dataframe)

        return table

    @staticmethod
    def from_iceberg(table: object) -> Table:
        """Creates a Daft Table instance from an Iceberg table.

        Args:
            table (object): a pyiceberg table

        Returns:
            Table: new daft table instance
        """
        try:
            from daft.catalog.__iceberg import IcebergTable

            return IcebergTable._from_obj(table)
        except ImportError:
            raise ImportError("Iceberg support not installed: pip install -U 'daft[iceberg]'")

    @staticmethod
    def from_unity(table: object) -> Table:
        """Returns a Daft Table instance from a Unity table.

        Args:
            table (object): unity table instance.
        """
        try:
            from daft.catalog.__unity import UnityTable

            return UnityTable._from_obj(table)
        except ImportError:
            raise ImportError("Unity support not installed: pip install -U 'daft[unity]'")

    @staticmethod
    def _from_obj(obj: object) -> Table:
        for factory in (Table.from_iceberg, Table.from_unity):
            try:
                return factory(obj)
            except ValueError:
                pass
            except ImportError:
                pass
        raise ValueError(
            f"Unsupported table type: {type(obj)}; please ensure all required extra dependencies are installed."
        )

    @staticmethod
    def _validate_options(method: str, input: dict[str, Any], valid: set[str]) -> None:
        """Validates input options against a set of valid options.

        Args:
            method (str): The method name to include in the error message
            input (dict[str, Any]): The input options dictionary
            valid (set[str]): Set of valid option keys

        Raises:
            ValueError: If any input options are not in the valid set
        """
        invalid_options = set(input.keys()) - valid
        if invalid_options:
            raise ValueError(f"Unsupported option(s) for {method}, found {invalid_options!s} not in {valid!s}")

    ###
    # read methods
    ###

    @abstractmethod
    def read(self, **options: Any) -> DataFrame:
        """Creates a new DataFrame from this table.

        Args:
            **options (Any): additional format-dependent read options

        Returns:
            DataFrame: new DataFrame instance
        """

    def select(self, *columns: ColumnInputType) -> DataFrame:
        """Creates a new DataFrame from the table applying the provided expressions.

        Args:
            *columns (Expression|str): columns to select from the current DataFrame

        Returns:
            DataFrame: new DataFrame instance with the select columns
        """
        return self.read().select(*columns)

    def show(self, n: int = 8) -> None:
        """Shows the first n rows from this table.

        Args:
            n (int): number of rows to show

        Returns:
            None
        """
        return self.read().show(n)

    ###
    # write methods
    ###

    def write(self, df: DataFrame, mode: Literal["append", "overwrite"] = "append", **options: Any) -> None:
        """Writes the DataFrame to this table.

        Args:
            df (DataFrame): datafram to write
            mode (str): write mode such as 'append' or 'overwrite'
            **options (Any): additional format-dependent write options
        """
        if mode == "append":
            return self.append(df, **options)

    @abstractmethod
    def append(self, df: DataFrame, **options: Any) -> None:
        """Appends the DataFrame to this table.

        Args:
            df (DataFrame): dataframe to append
            **options (Any): additional format-dependent write options
        """

    @abstractmethod
    def overwrite(self, df: DataFrame, **options: Any) -> None:
        """Overwrites this table with the given DataFrame.

        Args:
            df (DataFrame): dataframe to overwrite this table with
            **options (Any): additional format-dependent write options
        """

    ###
    # python methods
    ###

    def __repr__(self) -> str:
        return f"Table('{self.name}')"
